#!usr/bin/python
# -*- coding: utf-8-*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Aplicativo AdHoc Para generación de Facturas Electrónicas"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2009-2017 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.31b"

from datetime import datetime
from decimal import Decimal, getcontext, ROUND_DOWN
import os
import sys
import wx
import gui
import unicodedata
import traceback
from configparser import SafeConfigParser
from . import wsaa, wsfev1, wsfexv1
from .utils import SimpleXMLElement, SoapClient, SoapFault, date
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from smtplib import SMTP

#from PyFPDF.ejemplos.form import Form
from .pyfepdf import FEPDF

# Formatos de archivos:
from .formatos import formato_xml, formato_csv, formato_dbf, formato_txt, formato_json

try:
    from numeros import conv_text
except BaseException:
    def conv_text(num): return str(num)


HOMO = False
DEBUG = '--debug' in sys.argv
CONFIG_FILE = "rece.ini"

ACERCA_DE = """
PyRece: Aplicativo AdHoc para generar Facturas Electrónicas
Copyright (C) 2008-2015 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional y descargas ver:
http://www.sistemasagiles.com.ar/
"""

INSTRUCTIVO = """
Forma de uso:

 * Examinar: para buscar el archivo a procesar (opcional)
 * Cargar: para leer los datos del archivo de facturas a procesar
 * Autenticar: para iniciar la sesión en los servidores de AFIP (obligatorio antes de autorizar)
 * Marcar Todo: para seleccionar todas las facturas
 * Autorizar: para autorizar las facturas seleccionadas, completando el CAE y demás datos
 * Autorizar Lote: para autorizar en un solo lote las facturas seleccionadas
 * Grabar: para almacenar los datos procesados en el archivo de facturas
 * Previsualizar: para ver por pantalla la factura seleccionadas
 * Enviar: para envia por correo electrónico las facturas seleccionadas

Para solicitar soporte comercial, escriba a pyrece@sistemasagiles.com.ar
"""


class PyRece(gui.Controller):

    def on_load(self, event):
        self.cols = []
        self.items = []
        self.paths = [entrada]
        self.token = self.sign = ""
        self.smtp = None
        self.webservice = None
        if entrada and os.path.exists(entrada):
            self.cargar()

        self.components.cboWebservice.value = DEFAULT_WEBSERVICE
        self.on_cboWebservice_click(event)

        self.tipos = {
            1: "Factura A",
            2: "Notas de Débito A",
            3: "Notas de Crédito A",
            4: "Recibos A",
            5: "Notas de Venta al contado A",
            6: "Facturas B",
            7: "Notas de Débito B",
            8: "Notas de Crédito B",
            9: "Recibos B",
            10: "Notas de Venta al contado B",
            19: "Facturas de Exportación",
            20: "Nota de Débito por Operaciones con el Exterior",
            21: "Nota de Crédito por Operaciones con el Exterior",
            39: "Otros comprobantes A que cumplan con la R.G. N° 3419",
            40: "Otros comprobantes B que cumplan con la R.G. N° 3419",
            60: "Cuenta de Venta y Líquido producto A",
            61: "Cuenta de Venta y Líquido producto B",
            63: "Liquidación A",
            64: "Liquidación B",
            11: "Factura C",
            12: "Nota de Débito C",
            13: "Nota de Crédito C",
            15: "Recibo C",
        }

        self.component.bgcolor = "light gray"
        # deshabilito ordenar
        ##self.components.lvwListado.GetColumnSorter = lambda: lambda x,y: 0

    def set_cols(self, cols):
        self.__cols = cols
        lv = self.components.lvwListado
        # remove old columns:
        lv.clear_all()
        # insert new columns
        for col in cols:
            ch = gui.ListColumn(lv, name=col, text=col.replace("_", " ").title(), align="left")

    def get_cols(self):
        return self.__cols
    cols = property(get_cols, set_cols)

    def set_items(self, items):
        cols = self.cols
        self.__items = items

        def convert_str(value):
            if value is None:
                return ''
            elif isinstance(value, str):
                return str(value, 'latin1')
            elif isinstance(value, str):
                return value
            else:
                return str(value)
        self.components.lvwListado.items = [[convert_str(item[col]) for col in cols] for item in items]
        wx.SafeYield()

    def get_items(self):
        return self.__items
    items = property(get_items, set_items)

    def get_selected_items(self):
        for it in self.components.lvwListado.get_selected_items():
            yield it.index, it

    def set_selected_items(self, selected):
        for it in selected:
            it.selected = True

    def set_paths(self, paths):
        self.__paths = paths
        self.components.txtArchivo.value = ', '.join([fn for fn in paths])

    def get_paths(self):
        return self.__paths
    paths = property(get_paths, set_paths)

    def log(self, msg):
        if not isinstance(msg, str):
            msg = str(msg, "latin1", "ignore")
        print("LOG", msg)
        self.components.txtEstado.value = msg + "\n" + self.components.txtEstado.value
        wx.SafeYield()
        f = None
        try:
            f = open("pyrece.log", "a")
            f.write("%s: " % (datetime.now(), ))
            f.write(msg.encode("ascii", "ignore"))
            f.write("\n\r")
        except Exception as e:
            print(e)
        finally:
            if f:
                f.close()

    def progreso(self, value):
        if self.items:
            per = (value + 1) / float(len(self.items)) * 100
            self.components.pbProgreso.value = per
            wx.SafeYield()

    def error(self, code, text):
        ex = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
        self.log(''.join(ex))
        gui.alert(text, 'Error %s' % code)

    def verifica_ws(self):
        if not self.ws:
            gui.alert("Debe seleccionar el webservice a utilizar!", 'Advertencia')
            raise RuntimeError()
        if not self.token or not self.sign:
            gui.alert("Debe autenticarse con AFIP!", 'Advertencia')
            raise RuntimeError()
        self.ws.Dummy()

    def on_btnMarcarTodo_click(self, event):
        for it in self.components.lvwListado.items:
            it.selected = True

    def on_menu_consultas_dummy_click(self, event):
        # self.verifica_ws()
        try:
            if self.webservice in ("wsfev1", "wsfexv1"):
                self.ws.Dummy()
                msg = "AppServ %s\nDbServer %s\nAuthServer %s" % (
                    self.ws.AppServerStatus, self.ws.DbServerStatus, self.ws.AuthServerStatus)
                location = self.ws.client.location
            else:
                msg = "%s no soportado" % self.webservice
                location = ""
            gui.alert(msg, location)
        except Exception as e:
            self.error('Excepción', str(str(e), "latin1", "ignore"))

    def on_menu_consultas_lastCBTE_click(self, event):
        # self.verifica_ws()
        options = [v for k, v in sorted([(k, v) for k, v in list(self.tipos.items())])]
        result = gui.single_choice(options, "Tipo de comprobante",
                                   "Consulta Último Nro. Comprobante",
                                   )
        if not result:
            return
        tipocbte = [k for k, v in list(self.tipos.items()) if v == result][0]
        result = gui.prompt("Punto de venta",
                            "Consulta Último Nro. Comprobante", '2')
        if not result:
            return
        ptovta = result

        try:
            if self.webservice == "wsfev1":
                ultcmp = "%s (wsfev1)" % self.ws.CompUltimoAutorizado(tipocbte, ptovta)
            elif self.webservice == "wsfexv1":
                ultcmp = "%s (wsfexv1)" % self.ws.GetLastCMP(tipocbte, ptovta)

            gui.alert("Último comprobante: %s\n"
                      "Tipo: %s (%s)\nPunto de Venta: %s" % (ultcmp, self.tipos[tipocbte],
                                                             tipocbte, ptovta), 'Consulta Último Nro. Comprobante')
        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        except Exception as e:
            self.error('Excepción', str(str(e), "latin1", "ignore"))

    def on_menu_consultas_getCAE_click(self, event):
        self.verifica_ws()
        options = [v for k, v in sorted([(k, v) for k, v in list(self.tipos.items())])]
        result = gui.single_choice(options, "Tipo de comprobante",
                                   "Consulta Comprobante",
                                   )
        if not result:
            return
        tipocbte = [k for k, v in list(self.tipos.items()) if v == result][0]
        result = gui.prompt("Punto de venta",
                            "Consulta Comprobante", '2')
        if not result:
            return
        ptovta = result
        result = gui.prompt("Nº de comprobante",
                            "Consulta Comprobante", '2')
        if not result:
            return
        nrocbte = result

        try:
            if self.webservice == "wsfe":
                cae = 'no soportado'
            elif self.webservice == "wsfev1":
                cae = "%s (wsfev1)" % self.ws.CompConsultar(tipocbte, ptovta, nrocbte)
                self.log('CAE: %s' % self.ws.CAE)
                self.log('FechaCbte: %s' % self.ws.FechaCbte)
                self.log('PuntoVenta: %s' % self.ws.PuntoVenta)
                self.log('CbteNro: %s' % self.ws.CbteNro)
                self.log('ImpTotal: %s' % self.ws.ImpTotal)
                self.log('ImpNeto: %s' % self.ws.ImpNeto)
                self.log('ImptoLiq: %s' % self.ws.ImptoLiq)
                self.log('EmisionTipo: %s' % self.ws.EmisionTipo)
            elif self.webservice == "wsfexv1":
                cae = "%s (wsfexv1)" % self.ws.GetCMP(tipocbte, ptovta, nrocbte)
                self.log('CAE: %s' % self.ws.CAE)
                self.log('FechaCbte: %s' % self.ws.FechaCbte)
                self.log('PuntoVenta: %s' % self.ws.PuntoVenta)
                self.log('CbteNro: %s' % self.ws.CbteNro)
                self.log('ImpTotal: %s' % self.ws.ImpTotal)

            gui.alert("CAE: %s\n"
                      "Tipo: %s (%s)\nPunto de Venta: %s\nNumero: %s\nFecha: %s" % (
                          cae, self.tipos[tipocbte],
                          tipocbte, ptovta, nrocbte, self.ws.FechaCbte),
                      'Consulta Comprobante')

        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        except Exception as e:
            self.error('Excepción', str(str(e), "latin1", "ignore"))

    def on_menu_consultas_lastID_click(self, event):
        # self.verifica_ws()
        try:
            if self.webservice == "wsfexv1":
                ultnro = self.ws.GetLastID()
            else:
                ultnro = None
            gui.alert("Último ID (máximo): %s" % (ultnro),
                      'Consulta Último ID')
        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        except Exception as e:
            self.error('Excepción', str(e))

    def on_menu_ayuda_acercade_click(self, event):
        text = ACERCA_DE
        gui.alert(text, 'Acerca de PyRece Versión %s' % __version__)

    def on_menu_ayuda_instructivo_click(self, event):
        text = INSTRUCTIVO
        gui.alert(text, 'Instructivo de PyRece')

    def on_menu_ayuda_limpiar_click(self, event):
        self.components.txtEstado.value = ""

    def on_menu_ayuda_mensajesXML_click(self, event):
        self.verifica_ws()
        self.components.txtEstado.value = "XmlRequest:\n%s\n\nXmlResponse:\n%s" % (
            self.ws.xml_request, self.ws.xml_response)
        self.component.size = (592, 517)

    def on_menu_ayuda_estado_click(self, event):
        if self.component.size[1] < 517:
            self.component.size = (592, 517)
        else:
            self.component.size = (592, 265)

    def on_menu_ayuda_configuracion_click(self, event):
        self.components.txtEstado.value = open(CONFIG_FILE).read()
        self.component.size = (592, 517)

    def on_cboWebservice_click(self, event):
        self.webservice = self.components.cboWebservice.value
        self.ws = None
        self.token = None
        self.sign = None

        if self.webservice == "wsfev1":
            self.ws = wsfev1.WSFEv1()
        elif self.webservice == "wsfexv1":
            self.ws = wsfexv1.WSFEXv1()

    def on_btnAutenticar_click(self, event):
        try:
            if self.webservice in ('wsfe', ):
                service = "wsfe"
            elif self.webservice in ('wsfev1', ):
                self.log("Conectando WSFEv1... " + wsfev1_url)
                self.ws.Conectar("", wsfev1_url, proxy_dict, timeout=60, cacert=CACERT, wrapper=WRAPPER)
                self.ws.Cuit = cuit
                service = "wsfe"
            elif self.webservice in ('wsfex', 'wsfexv1'):
                self.log("Conectando WSFEXv1... " + wsfexv1_url)
                self.ws.Conectar("", wsfexv1_url, proxy_dict, cacert=CACERT, wrapper=WRAPPER)
                self.ws.Cuit = cuit
                service = "wsfex"
            else:
                gui.alert('Debe seleccionar servicio web!', 'Advertencia')
                return

            self.log("Creando TRA %s ..." % service)
            ws = wsaa.WSAA()
            tra = ws.CreateTRA(service)
            self.log("Frimando TRA (CMS) con %s %s..." % (str(cert), str(privatekey)))
            cms = ws.SignTRA(str(tra), str(cert), str(privatekey))
            self.log("Llamando a WSAA... " + wsaa_url)
            ws.Conectar("", wsdl=wsaa_url, proxy=proxy_dict, cacert=CACERT, wrapper=WRAPPER)
            self.log("Proxy: %s" % proxy_dict)
            xml = ws.LoginCMS(str(cms))
            self.log("Procesando respuesta...")
            if xml:
                self.token = ws.Token
                self.sign = ws.Sign
            if DEBUG:
                self.log("Token: %s" % self.token)
                self.log("Sign: %s" % self.sign)
            elif self.token and self.sign:
                self.log("Token: %s... OK" % self.token[:10])
                self.log("Sign: %s... OK" % self.sign[:10])
            if self.webservice in ("wsfev1", "wsfexv1"):
                self.ws.Token = self.token
                self.ws.Sign = self.sign

            if xml:
                gui.alert('Autenticado OK!', 'Advertencia')
            else:
                gui.alert('Respuesta: %s' % ws.XmlResponse, 'No se pudo autenticar: %s' % ws.Excepcion)
        except SoapFault as e:
            self.error(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        except Exception as e:
            self.error('Excepción', str(e))

    def examinar(self):
        filename = entrada
        wildcard = ["Planillas Excel (*.xlsx)|*.xlsx",
                    "Archivos CSV (*.csv)|*.csv",
                    "Archivos XML (*.xml)|*.xml",
                    "Archivos TXT (*.txt)|*.txt",
                    "Archivos DBF (*.dbf)|*.dbf",
                    "Archivos JSON (*.json)|*.json",
                    ]
        if entrada.endswith("xml"):
            wildcard.sort(reverse=True)

        result = gui.open_file('Abrir', 'datos', filename, '|'.join(wildcard))
        if not result:
            return
        self.paths = [result]

    def on_menu_archivo_abrir_click(self, event):
        self.examinar()
        self.cargar()

    def on_menu_archivo_cargar_click(self, event):
        self.cargar()

    def cargar(self):
        try:
            items = []
            for fn in self.paths:
                if fn.lower().endswith(".csv") or fn.lower().endswith(".xlsx"):
                    filas = formato_csv.leer(fn)
                    items.extend(filas)
                elif fn.lower().endswith(".xml"):
                    regs = formato_xml.leer(fn)
                    items.extend(formato_csv.aplanar(regs))
                elif fn.lower().endswith(".txt"):
                    regs = formato_txt.leer(fn)
                    items.extend(formato_csv.aplanar(regs))
                elif fn.lower().endswith(".dbf"):
                    reg = formato_dbf.leer(conf_dbf, carpeta=os.path.dirname(fn))
                    items.extend(formato_csv.aplanar(list(reg.values())))
                elif fn.lower().endswith(".json"):
                    regs = formato_json.leer(fn)
                    items.extend(formato_csv.aplanar(regs))
                else:
                    self.error('Formato de archivo desconocido: %s', str(fn))
            if len(items) < 2:
                gui.alert('El archivo no tiene datos válidos', 'Advertencia')
            # extraer los nombres de columnas (ignorar vacios de XLSX)
            cols = items and [str(it).strip() for it in items[0] if it] or []
            if DEBUG:
                print("Cols", cols)
            # armar diccionario por cada linea
            items = [dict([(col, item[i]) for i, col in enumerate(cols)])
                     for item in items[1:]]
            self.cols = cols
            self.items = items
        except Exception as e:
            self.error('Excepción', str(e))
            # raise

    def on_menu_archivo_guardar_click(self, event):
        filename = entrada
        wildcard = ["Archivos CSV (*.csv)|*.csv", "Archivos XML (*.xml)|*.xml",
                    "Archivos TXT (*.txt)|*.txt", "Archivos DBF (*.dbf)|*.dbf",
                    "Archivos JSON (*.json)|*.json",
                    "Planillas Excel (*.xlsx)|*.xlsx",
                    ]
        if entrada.endswith("xml"):
            wildcard.sort(reverse=True)
        if self.paths:
            path = self.paths[0]
        else:
            path = salida
        result = gui.save_file(title='Guardar', filename=path,
                               wildcard='|'.join(wildcard))
        if not result:
            return
        fn = result[0]
        self.grabar(fn)

    def grabar(self, fn=None):
        try:
            if fn is None and salida:
                if salida.startswith("-") and self.paths:
                    fn = os.path.splitext(self.paths[0])[0] + salida
                else:
                    fn = salida
            elif not fn:
                raise RuntimeError("Debe indicar un nombre de archivo para grabar")
            if fn.lower().endswith(".csv") or fn.lower().endswith(".xlsx"):
                formato_csv.escribir([self.cols] + [[item[k] for k in self.cols] for item in self.items], fn)
            else:
                regs = formato_csv.desaplanar([self.cols] + [[item[k] for k in self.cols] for item in self.items])
                if fn.endswith(".xml"):
                    formato_xml.escribir(regs, fn)
                elif fn.endswith(".txt"):
                    formato_txt.escribir(regs, fn)
                elif fn.endswith(".dbf"):
                    formato_dbf.escribir(regs, conf_dbf, carpeta=os.path.dirname(fn))
                elif fn.endswith(".json"):
                    formato_json.escribir(regs, fn)
                else:
                    self.error('Formato de archivo desconocido', str(fn))
            gui.alert('Se guardó con éxito el archivo:\n%s' % (str(fn),), 'Guardar')
        except Exception as e:
            self.error('Excepción', str(e))

    def on_btnAutorizar_click(self, event):
        self.verifica_ws()
        try:
            ok = procesadas = rechazadas = 0
            cols = self.cols
            items = []
            self.progreso(0)
            selected = []
            for i, item in self.get_selected_items():
                kargs = item.copy()
                selected.append(item)
                kargs['cbt_desde'] = kargs['cbt_hasta'] = kargs['cbt_numero']
                for key in kargs:
                    if isinstance(kargs[key], str):
                        kargs[key] = kargs[key].replace(",", ".")
                if self.webservice == 'wsfev1':
                    encabezado = {}
                    for k in ('concepto', 'tipo_doc', 'nro_doc', 'tipo_cbte', 'punto_vta',
                              'cbt_desde', 'cbt_hasta', 'imp_total', 'imp_tot_conc', 'imp_neto',
                              'imp_iva', 'imp_trib', 'imp_op_ex', 'fecha_cbte',
                              'moneda_id', 'moneda_ctz'):
                        encabezado[k] = kargs[k]

                    for k in ('fecha_venc_pago', 'fecha_serv_desde', 'fecha_serv_hasta'):
                        if k in kargs:
                            encabezado[k] = kargs.get(k)

                    self.ws.CrearFactura(**encabezado)

                    for l in range(1, 1000):
                        k = 'tributo_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            desc = kargs[k % 'desc']
                            base_imp = kargs[k % 'base_imp']
                            alic = kargs[k % 'alic']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarTributo(id, desc, base_imp, alic, importe)
                        else:
                            break

                    for l in range(1, 1000):
                        k = 'iva_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            base_imp = kargs[k % 'base_imp']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarIva(id, base_imp, importe)
                        else:
                            break

                    for l in range(1, 1000):
                        k = 'cbte_asoc_%%s_%s' % l
                        if (k % 'tipo') in kargs:
                            tipo = kargs[k % 'tipo']
                            pto_vta = kargs[k % 'pto_vta']
                            nro = kargs[k % 'nro']
                            if id:
                                self.ws.AgregarCmpAsoc(tipo, pto_vta, nro)
                        else:
                            break

                    for l in range(1, 1000):
                        k = 'opcional_%%s_%s' % l
                        if (k % 'id') in kargs:
                            op_id = kargs[k % 'id']
                            valor = kargs[k % 'valor']
                            if op_id:
                                self.ws.AgregarOpcional(op_id, valor)
                        else:
                            break

                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k, v) for k, v in list(self.ws.factura.items())]))

                    cae = self.ws.CAESolicitar()
                    kargs.update({
                        'cae': self.ws.CAE,
                        'fecha_vto': self.ws.Vencimiento,
                        'resultado': self.ws.Resultado,
                        'motivo': self.ws.Obs,
                        'reproceso': self.ws.Reproceso,
                        'err_code': self.ws.ErrCode.encode("latin1"),
                        'err_msg': self.ws.ErrMsg.encode("latin1"),
                    })
                    if self.ws.ErrMsg:
                        gui.alert(self.ws.ErrMsg, "Error AFIP")
                    if self.ws.Obs and self.ws.Obs != '00':
                        gui.alert(self.ws.Obs, "Observación AFIP")

                elif self.webservice == 'wsfexv1':
                    kargs['cbte_nro'] = kargs['cbt_numero']
                    kargs['permiso_existente'] = kargs['permiso_existente'] or ""
                    encabezado = {}
                    for k in ('tipo_cbte', 'punto_vta', 'cbte_nro', 'fecha_cbte',
                              'imp_total', 'tipo_expo', 'permiso_existente', 'pais_dst_cmp',
                              'nombre_cliente', 'cuit_pais_cliente', 'domicilio_cliente',
                              'id_impositivo', 'moneda_id', 'moneda_ctz',
                              'obs_comerciales', 'obs_generales', 'forma_pago', 'incoterms',
                              'idioma_cbte', 'incoterms_ds'):
                        encabezado[k] = kargs.get(k)

                    self.ws.CrearFactura(**encabezado)

                    for l in range(1, 1000):
                        k = 'codigo%s' % l
                        if k in kargs:
                            codigo = kargs['codigo%s' % l]
                            ds = kargs['descripcion%s' % l]
                            qty = kargs['cantidad%s' % l]
                            umed = kargs['umed%s' % l]
                            precio = kargs['precio%s' % l]
                            importe = kargs['importe%s' % l]
                            bonif = kargs.get('bonif%s' % l)
                            self.ws.AgregarItem(codigo, ds, qty, umed, precio, importe, bonif)
                        else:
                            break

                    for l in range(1, 1000):
                        k = 'cbte_asoc_%%s_%s' % l
                        if (k % 'tipo') in kargs:
                            tipo = kargs[k % 'tipo']
                            pto_vta = kargs[k % 'pto_vta']
                            nro = kargs[k % 'nro']
                            if id:
                                self.ws.AgregarCmpAsoc(tipo, pto_vta, nro)
                        else:
                            break

                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k, v) for k, v in list(self.ws.factura.items())]))

                    cae = self.ws.Authorize(kargs['id'])
                    kargs.update({
                        'cae': self.ws.CAE,
                        'fecha_vto': self.ws.Vencimiento,
                        'resultado': self.ws.Resultado,
                        'motivo': self.ws.Obs,
                        'reproceso': self.ws.Reproceso,
                        'err_code': self.ws.ErrCode.encode("latin1"),
                        'err_msg': self.ws.ErrMsg.encode("latin1"),
                    })
                    if self.ws.ErrMsg:
                        gui.alert(self.ws.ErrMsg, "Error AFIP")
                    if self.ws.Obs and self.ws.Obs != '00':
                        gui.alert(self.ws.Obs, "Observación AFIP")

                # actualizo la factura
                for k in ('cae', 'fecha_vto', 'resultado', 'motivo', 'reproceso', 'err_code', 'err_msg'):
                    if kargs.get(k):
                        item[k] = kargs[k] if kargs[k] is not None else ""
                self.items[i] = item
                self.log("ID: %s CAE: %s Motivo: %s Reproceso: %s" % (kargs['id'], kargs['cae'], kargs['motivo'], kargs['reproceso']))
                procesadas += 1
                if kargs['resultado'] == "R":
                    rechazadas += 1
                elif kargs['resultado'] == "A":
                    ok += 1
                self.progreso(i)
            self.items = self.items
            self.set_selected_items(selected)
            self.progreso(len(self.items) - 1)
            gui.alert('Proceso finalizado, procesadas %d\n\n'
                      'Aceptadas: %d\n'
                      'Rechazadas: %d' % (procesadas, ok, rechazadas),
                      'Autorización')
            self.grabar()
        except SoapFault as e:
            self.error(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        except KeyError as e:
            self.error("Error", 'Campo obligatorio no encontrado: %s' % e)
        except Exception as e:
            self.error('Excepción', str(e))
        finally:
            if DEBUG:
                if self.webservice == 'wsfev1' and DEBUG:
                    print(self.ws.XmlRequest)
                    print(self.ws.XmlResponse)

    def on_btnAutorizarLote_click(self, event):
        self.verifica_ws()
        if not self.items:
            return
        try:
            #getcontext().prec = 2
            ok = 0
            rechazadas = 0
            cols = self.cols
            items = []
            self.progreso(0)
            cbt_desde = cbt_hasta = None
            datos = {
                'tipo_cbte': None,
                'punto_vta': None,
                'fecha_cbte': None,
                'fecha_venc_pago': None,
                'fecha_cbte': None,
                'fecha_venc_pago': None,
                'fecha_serv_desde': None,
                'fecha_serv_hasta': None,
                'moneda_id': None,
                'moneda_ctz': None,
                'id': None,
            }
            importes = {
                'imp_total': Decimal(0),
                'imp_tot_conc': Decimal(0),
                'imp_neto': Decimal(0),
                'imp_iva': Decimal(0),
                'imp_op_ex': Decimal(0),
                'imp_trib': Decimal(0),
            }
            for l in range(1, 5):
                k = 'iva_%%s_%s' % l
                datos[k % 'id'] = None
                importes[k % 'base_imp'] = Decimal(0)
                importes[k % 'importe'] = Decimal(0)

            for l in range(1, 10):
                k = 'tributo_%%s_%s' % l
                datos[k % 'id'] = None
                datos[k % 'desc'] = None
                importes[k % 'base_imp'] = Decimal(0)
                datos[k % 'alic'] = None
                importes[k % 'importe'] = Decimal(0)

            for i, item in self.get_selected_items():
                if cbt_desde is None or int(item['cbt_numero']) < cbt_desde:
                    cbt_desde = int(item['cbt_numero'])
                if cbt_hasta is None or int(item['cbt_numero']) > cbt_hasta:
                    cbt_hasta = int(item['cbt_numero'])
                for key in item:
                    if key in datos:
                        if datos[key] is None:
                            datos[key] = item[key]
                        elif datos[key] != item[key]:
                            raise RuntimeError("%s tiene valores distintos en el lote!" % key)
                    if key in importes and item[key]:
                        importes[key] = importes[key] + Decimal("%.2f" % float(str(item[key].replace(",", "."))))

            kargs = {'cbt_desde': cbt_desde, 'cbt_hasta': cbt_hasta}
            kargs.update({'tipo_doc': 99, 'nro_doc': '0'})
            kargs.update(datos)
            kargs.update(importes)
            if kargs['fecha_serv_desde'] and kargs['fecha_serv_hasta']:
                kargs['presta_serv'] = 1
                kargs['concepto'] = 2
            else:
                kargs['presta_serv'] = 0
                kargs['concepto'] = 1
                del kargs['fecha_serv_desde']
                del kargs['fecha_serv_hasta']

            for key, val in list(importes.items()):
                importes[key] = val.quantize(Decimal('.01'), rounding=ROUND_DOWN)

            if 'id' not in kargs or kargs['id'] == "":
                id = int(kargs['cbt_desde'])
                id += (int(kargs['tipo_cbte']) * 10**4 + int(kargs['punto_vta'])) * 10**8
                kargs['id'] = id

            if DEBUG:
                self.log('\n'.join(["%s='%s'" % (k, v) for k, v in list(kargs.items())]))
            if '--test' in sys.argv:
                kargs['cbt_desde'] = 777
                kargs['fecha_cbte'] = '20110802'
                kargs['fecha_venc_pago'] = '20110831'

            if gui.confirm("Confirma Lote:\n"
                           "Tipo: %(tipo_cbte)s Desde: %(cbt_desde)s Hasta %(cbt_hasta)s\n"
                           "Neto: %(imp_neto)s IVA: %(imp_iva)s Trib.: %(imp_trib)s Total: %(imp_total)s"
                           % kargs, "Autorizar lote:"):

                if self.webservice == 'wsfev1':
                    encabezado = {}
                    for k in ('concepto', 'tipo_doc', 'nro_doc', 'tipo_cbte', 'punto_vta',
                              'cbt_desde', 'cbt_hasta', 'imp_total', 'imp_tot_conc', 'imp_neto',
                              'imp_iva', 'imp_trib', 'imp_op_ex', 'fecha_cbte',
                              'moneda_id', 'moneda_ctz'):
                        encabezado[k] = kargs[k]

                    for k in ('fecha_venc_pago', 'fecha_serv_desde', 'fecha_serv_hasta'):
                        if k in kargs:
                            encabezado[k] = kargs.get(k)

                    self.ws.CrearFactura(**encabezado)
                    for l in range(1, 1000):
                        k = 'iva_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            base_imp = kargs[k % 'base_imp']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarIva(id, base_imp, importe)
                        else:
                            break

                    for l in range(1, 1000):
                        k = 'tributo_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            desc = kargs[k % 'desc']
                            base_imp = kargs[k % 'base_imp']
                            alic = kargs[k % 'alic']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarTributo(id, desc, base_imp, alic, importe)
                        else:
                            break

                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k, v) for k, v in list(self.ws.factura.items())]))

                    cae = self.ws.CAESolicitar()
                    kargs.update({
                        'cae': self.ws.CAE,
                        'fecha_vto': self.ws.Vencimiento,
                        'resultado': self.ws.Resultado,
                        'motivo': self.ws.Obs,
                        'reproceso': self.ws.Reproceso,
                        'err_code': self.ws.ErrCode.encode("latin1"),
                        'err_msg': self.ws.ErrMsg.encode("latin1"),
                    })
                    if self.ws.ErrMsg:
                        gui.alert(self.ws.ErrMsg, "Error AFIP")
                    if self.ws.Obs and self.ws.Obs != '00':
                        gui.alert(self.ws.Obs, "Observación AFIP")

                for i, item in self.get_selected_items():
                    for key in ('id', 'cae', 'fecha_vto', 'resultado', 'motivo', 'reproceso', 'err_code', 'err_msg'):
                        item[key] = kargs[key] if kargs[key] is not None else ""
                        self.items[i] = item

                self.log("ID: %s CAE: %s Motivo: %s Reproceso: %s" % (kargs['id'], kargs['cae'], kargs['motivo'], kargs['reproceso']))
                if kargs['resultado'] == "R":
                    rechazadas += 1
                elif kargs['resultado'] == "A":
                    ok += 1

                self.items = self.items  # refrescar, ver de corregir
                self.progreso(len(self.items))
                gui.alert('Proceso finalizado OK!\n\nAceptadas: %d\nRechazadas: %d' % (ok, rechazadas), 'Autorización')
                self.grabar()
        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii", "ignore"))
        except Exception as e:
            self.error('Excepción', str(e))

    def on_btnPrevisualizar_click(self, event):
        try:
            j = 0
            for i, item in self.get_selected_items():
                j += 1
                archivo = self.generar_factura(item, mostrar=(j == 1))
        except Exception as e:
            print(e)
            self.error('Excepción', str(str(e), 'latin1', 'ignore'))

    def on_btnEnviar_click(self, event):
        try:
            ok = no = 0
            self.progreso(0)
            for i, item in self.get_selected_items():
                if not item['cae'] in ("", "NULL"):
                    archivo = self.generar_factura(item)
                    if item.get('email'):
                        self.enviar_mail(item, archivo)
                        ok += 1
                    else:
                        no += 1
                        self.log("No se envia factura %s por no tener EMAIL" % item['cbt_numero'])
                else:
                    self.log("No se envia factura %s por no tener CAE" % item['cbt_numero'])
                    no += 1
                self.progreso(i)
            self.progreso(len(self.items))
            gui.alert('Proceso finalizado OK!\n\nEnviados: %d\nNo enviados: %d' % (ok, no), 'Envio de Email')
        except Exception as e:
            self.error('Excepción', str(e))

    def generar_factura(self, fila, mostrar=False):

        fepdf = FEPDF()
        fact = formato_csv.desaplanar([self.cols] + [[item[k] for k in self.cols] for item in [fila]])[0]
        fact['cbte_nro'] = fact['cbt_numero']
        fact['items'] = fact['detalles']

        for d in fact['datos']:
            fepdf.AgregarDato(d['campo'], d['valor'], d['pagina'])
            # por compatiblidad, completo campos anteriores
            if d['campo'] not in fact and d['valor']:
                fact[d['campo']] = d['valor']

        fepdf.factura = fact

        # convertir importe total en texto (palabras):
        moneda_ds = {"PES": "PESOS", "DOL": "DOLAR EEUU"}.get(fact.get("moneda_id", ""), "")
        fact["en_letras"] = "SON " + moneda_ds + " " + conv_text(float(fact["imp_total"]))

        # cargo el formato CSV por defecto (factura.csv)
        fepdf.CargarFormato(conf_fact.get("formato", "factura.csv"))

        # establezco formatos (cantidad de decimales) según configuración:
        fepdf.FmtCantidad = conf_fact.get("fmt_cantidad", "0.2")
        fepdf.FmtPrecio = conf_fact.get("fmt_precio", "0.2")

        # datos fijos:
        fepdf.CUIT = cuit  # CUIT del emisor para código de barras
        for k, v in list(conf_pdf.items()):
            fepdf.AgregarDato(k, v)

        fepdf.CrearPlantilla(papel=conf_fact.get("papel", "legal"),
                             orientacion=conf_fact.get("orientacion", "portrait"))
        fepdf.ProcesarPlantilla(num_copias=int(conf_fact.get("copias", 1)),
                                lineas_max=int(conf_fact.get("lineas_max", 24)),
                                qty_pos=conf_fact.get("cant_pos") or 'izq')

        salida = conf_fact.get("salida", "")
        fact = fepdf.factura

        if salida:
            pass
        elif 'pdf' in fact and fact['pdf']:
            salida = fact['pdf']
        else:
            # genero el nombre de archivo según datos de factura
            d = conf_fact.get('directorio', ".")
            clave_subdir = conf_fact.get('subdirectorio', 'fecha_cbte')
            if clave_subdir:
                d = os.path.join(d, item[clave_subdir])
            if not os.path.isdir(d):
                os.mkdir(d)
            fs = conf_fact.get('archivo', 'numero').split(",")
            it = item.copy()
            tipo_fact, letra_fact, numero_fact = fact['_fmt_fact']
            it['tipo'] = tipo_fact.replace(" ", "_")
            it['letra'] = letra_fact
            it['numero'] = numero_fact
            it['mes'] = item['fecha_cbte'][4:6]
            it['año'] = item['fecha_cbte'][0:4]
            # remover acentos, ñ del nombre de archivo (vía unicode):
            fn = ''.join([str(it.get(ff, ff)) for ff in fs])
            fn = unicodedata.normalize('NFKD', fn).encode('ASCII', 'ignore')
            salida = os.path.join(d, "%s.pdf" % fn)
        fepdf.GenerarPDF(archivo=salida)
        if mostrar:
            fepdf.MostrarPDF(archivo=salida, imprimir='--imprimir' in sys.argv)

        return salida

    def enviar_mail(self, item, archivo):
        archivo = self.generar_factura(item)
        if item['email']:
            msg = MIMEMultipart()
            msg['Subject'] = conf_mail['motivo'].replace("NUMERO", str(item['cbt_numero']))
            msg['From'] = conf_mail['remitente']
            msg['Reply-to'] = msg['From']
            msg['To'] = item['email']
            msg.preamble = 'Mensaje de multiples partes.\n'
            if not 'html' in conf_mail:
                part = MIMEText(conf_mail['cuerpo'])
                msg.attach(part)
            else:
                alt = MIMEMultipart('alternative')
                msg.attach(alt)
                text = MIMEText(conf_mail['cuerpo'])
                alt.attach(text)
                # We reference the image in the IMG SRC attribute by the ID we give it below
                html = MIMEText(conf_mail['html'], 'html')
                alt.attach(html)
            part = MIMEApplication(open(archivo, "rb").read())
            part.add_header('Content-Disposition', 'attachment', filename=os.path.basename(archivo))
            msg.attach(part)

            try:
                self.log("Enviando email: %s a %s" % (msg['Subject'], msg['To']))
                if not self.smtp:
                    self.smtp = SMTP(conf_mail['servidor'], conf_mail.get('puerto', 25))
                    if conf_mail['usuario'] and conf_mail['clave']:
                        self.smtp.ehlo()
                        if conf_mail.get('tls', False):
                            self.log("Iniciando TLS...")
                            self.smtp.starttls()
                            self.smtp.ehlo()
                        self.smtp.login(conf_mail['usuario'], conf_mail['clave'])
                to = [msg['To']]
                bcc = conf_mail.get('bcc', None)
                if bcc:
                    to.append(bcc)
                self.smtp.sendmail(msg['From'], to, msg.as_string())
            except Exception as e:
                self.error('Excepción', str(e))


if __name__ == '__main__':

    if len(sys.argv) > 1 and not sys.argv[1].startswith("-"):
        CONFIG_FILE = sys.argv[1]
    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    if not len(config.sections()):
        if os.path.exists(CONFIG_FILE):
            gui.alert("Error al cargar archivo de configuración: %s" %
                      CONFIG_FILE, "PyRece: Imposible Continuar")
        else:
            gui.alert("No se encuentra archivo de configuración: %s" %
                      CONFIG_FILE, "PyRece: Imposible Continuar")
        sys.exit(1)
    cert = config.get('WSAA', 'CERT')
    privatekey = config.get('WSAA', 'PRIVATEKEY')
    cuit = config.get('WSFEv1', 'CUIT')
    if config.has_option('WSFEv1', 'ENTRADA'):
        entrada = config.get('WSFEv1', 'ENTRADA')
    else:
        entrada = ""
    if not os.path.exists(entrada):
        entrada = "facturas.csv"
    if config.has_option('WSFEv1', 'SALIDA'):
        salida = config.get('WSFEv1', 'SALIDA')
    else:
        salida = "resultado.csv"

    if config.has_section('FACTURA'):
        conf_fact = dict(config.items('FACTURA'))
    else:
        conf_fact = {}

    conf_pdf = dict(config.items('PDF'))
    conf_mail = dict(config.items('MAIL'))

    if config.has_section('DBF'):
        conf_dbf = dict(config.items('DBF'))
    else:
        conf_dbf = {}

    if config.has_option('WSAA', 'URL') and not HOMO:
        wsaa_url = config.get('WSAA', 'URL')
    else:
        wsaa_url = wsaa.WSAAURL

    if config.has_option('WSFEv1', 'URL') and not HOMO:
        wsfev1_url = config.get('WSFEv1', 'URL')
    else:
        wsfev1_url = wsfev1.WSDL

    if config.has_option('WSFEXv1', 'URL') and not HOMO:
        wsfexv1_url = config.get('WSFEXv1', 'URL')
    else:
        wsfexv1_url = wsfexv1.WSDL

    CACERT = config.has_option('WSAA', 'CACERT') and config.get('WSAA', 'CACERT') or None
    WRAPPER = config.has_option('WSAA', 'WRAPPER') and config.get('WSAA', 'WRAPPER') or None

    DEFAULT_WEBSERVICE = "wsfev1"
    if config.has_section('PYRECE'):
        DEFAULT_WEBSERVICE = config.get('PYRECE', 'WEBSERVICE')

    if config.has_section('PROXY'):
        proxy_dict = dict(("proxy_%s" % k, v) for k, v in config.items('PROXY'))
        proxy_dict['proxy_port'] = int(proxy_dict['proxy_port'])
    else:
        proxy_dict = {}

    c = PyRece()
    gui.main_loop()



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para obtener código de autorización electrónico CAE webservice
WSCT de AFIP (Factura Electrónica Comprobantes de Turismo)
Resolución Conjunta General 3971 y Resolución 566/2016.
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2017 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.02c"

import datetime
import decimal
import os
import sys
from .utils import verifica, inicializar_y_capturar_excepciones, BaseWS, get_install_dir

HOMO = False
LANZAR_EXCEPCIONES = True
WSDL = "https://fwshomo.afip.gov.ar/wsct/CTService?wsdl"


class WSCT(BaseWS):
    "Interfaz para el WebService de Factura Electrónica Comprobantes Turismo"
    _public_methods_ = ['CrearFactura', 'EstablecerCampoFactura', 'AgregarIva', 'AgregarItem',
                        'AgregarTributo', 'AgregarCmpAsoc', 'EstablecerCampoItem',
                        'AgregarDatoAdicional', 'AgregarFormaPago',
                        'AutorizarComprobante', 'CAESolicitar',
                        'InformarCAEANoUtilizado', 'InformarCAEANoUtilizadoPtoVta',
                        'ConsultarUltimoComprobanteAutorizado', 'CompUltimoAutorizado',
                        'ConsultarPtosVtaCAEANoInformados',
                        'ConsultarComprobante',
                        'ConsultarTiposComprobante', 'ConsultarTiposDocumento',
                        'consultarTiposIVA', 'ConsultarCondicionesIVA',
                        'ConsultarMonedas', 'ConsultarCotizacion',
                        'ConsultarTiposItem', 'ConsultarCodigosItemTurismo',
                        'ConsultarTiposTributo',
                        'ConsultarCUITsPaises', 'ConsultarPaises',
                        'ConsultarTiposDatosAdicionales', 'ConsultarFomasPago',
                        'ConsultarTiposTarjeta', 'ConsultarTiposCuenta',
                        'ConsultarPuntosVenta',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        'SetParametros', 'SetTicketAcceso', 'GetParametro',
                        'Dummy', 'Conectar', 'DebugLog', 'SetTicketAcceso']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'XmlRequest', 'XmlResponse', 'Version', 'InstallDir', 'LanzarExcepciones',
                      'Resultado', 'Obs', 'Observaciones', 'ErrCode', 'ErrMsg',
                      'EmisionTipo', 'Reproceso', 'Reprocesar', 'Evento',
                      'CAE', 'Vencimiento', 'Evento', 'Errores', 'Traceback', 'Excepcion',
                      'CAEA', 'Periodo', 'Orden', 'FchVigDesde', 'FchVigHasta', 'FchTopeInf', 'FchProceso',
                      'CbteNro', 'FechaCbte', 'PuntoVenta', 'ImpTotal']

    _reg_progid_ = "WSCT"
    _reg_clsid_ = "{5DE7917D-CE97-4C88-B6C7-DAF8CEB54E93}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
    Reprocesar = True  # recuperar automaticamente CAE emitidos
    LanzarExcepciones = LANZAR_EXCEPCIONES
    factura = None

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.Resultado = self.Motivo = self.Reproceso = ''
        self.LastID = self.LastCMP = self.CAE = self.Vencimiento = ''
        self.CAEA = None
        self.Periodo = self.Orden = ""
        self.FchVigDesde = self.FchVigHasta = ""
        self.FchTopeInf = self.FchProceso = ""
        self.CbteNro = self.FechaCbte = self.ImpTotal = None
        self.PuntoVenta = self.EmisionTipo = self.Evento = ''
        self.Reproceso = ''  # no implementado

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        for key in ('arrayErrores', 'arrayErroresFormato'):
            errores = ret.get(key, [])
            for error in errores:
                for k in ('codigoDescripcion', 'codigoDescripcionString'):
                    err = error.get(k)
                    if err:
                        break
                self.Errores.append("%s: %s" % (
                    err['codigo'],
                    err['descripcion'],
                ))
            self.ErrMsg = '\n'.join(self.Errores)

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        ret = self.client.dummy()
        result = ret['dummyReturn']
        self.AppServerStatus = result['appserver']
        self.DbServerStatus = result['dbserver']
        self.AuthServerStatus = result['authserver']
        return True

    def CrearFactura(self, tipo_doc=None, nro_doc=None, tipo_cbte=None, punto_vta=None,
                     cbte_nro=None, imp_total=None, imp_tot_conc=None, imp_neto=None,
                     imp_subtotal=None, imp_trib=None, imp_op_ex=None, imp_reintegro=None,
                     fecha_cbte=None, id_impositivo="", cod_pais=None, domicilio="", cod_relacion="",
                     moneda_id=None, moneda_ctz=None, observaciones=None,
                     **kwargs
                     ):
        "Creo un objeto factura (interna)"
        # Creo una factura electronica de exportación
        fact = {'tipo_doc': tipo_doc, 'nro_doc': nro_doc,
                'tipo_cbte': tipo_cbte, 'punto_vta': punto_vta,
                'cbte_nro': cbte_nro,
                'id_impositivo': id_impositivo,
                'cod_pais': cod_pais,
                'domicilio': domicilio,
                'cod_relacion': cod_relacion,
                'imp_total': imp_total, 'imp_tot_conc': imp_tot_conc,
                'imp_neto': imp_neto,
                'imp_subtotal': imp_subtotal,  # 'imp_iva': imp_iva,
                'imp_trib': imp_trib, 'imp_op_ex': imp_op_ex,
                'imp_reintegro': imp_reintegro,
                'fecha_cbte': fecha_cbte,
                'moneda_id': moneda_id, 'moneda_ctz': moneda_ctz,
                'observaciones': observaciones,
                'cbtes_asoc': [],
                'tributos': [],
                'iva': [],
                'detalles': [],
                'adicionales': [],
                'formas_pago': [],
                }

        self.factura = fact
        return True

    def EstablecerCampoFactura(self, campo, valor):
        if campo in self.factura or campo in ('fecha_serv_desde', 'fecha_serv_hasta', 'caea', 'fch_venc_cae'):
            self.factura[campo] = valor
            return True
        else:
            return False

    def AgregarCmpAsoc(self, tipo=1, pto_vta=0, nro=0, cuit=None, **kwargs):
        "Agrego un comprobante asociado a una factura (interna)"
        cmp_asoc = {
            'tipo': tipo,
            'pto_vta': pto_vta,
            'nro': nro}
        if cuit is not None:
            cmp_asoc['cuit'] = cuit
        self.factura['cbtes_asoc'].append(cmp_asoc)
        return True

    def AgregarTributo(self, tributo_id, desc, base_imp, alic, importe, **kwargs):
        "Agrego un tributo a una factura (interna)"
        tributo = {
            'tributo_id': tributo_id,
            'desc': desc,
            'base_imp': base_imp,
            'importe': importe,
        }
        self.factura['tributos'].append(tributo)
        return True

    def AgregarIva(self, iva_id, base_imp, importe, **kwargs):
        "Agrego un tributo a una factura (interna)"
        iva = {
            'iva_id': iva_id,
            'importe': importe,
        }
        self.factura['iva'].append(iva)
        return True

    def AgregarItem(self, tipo=None, cod_tur=None,
                    codigo=None, ds=None,
                    iva_id=None, imp_iva=None, imp_subtotal=None, **kwargs):
        "Agrego un item a una factura (interna)"
        # ds = unicode(ds, "latin1") # convierto a latin1
        # Nota: no se calcula neto, iva, etc (deben venir calculados!)
        tipo = int(tipo)
        if tipo == 99:
            imp_subtotal = -abs(float(imp_subtotal))
            imp_iva = -abs(float(imp_iva))
        item = {
            'tipo': tipo,
            'cod_tur': cod_tur,
            'codigo': codigo,
            'ds': ds,
            'iva_id': iva_id,
            'imp_iva': imp_iva,
            'imp_subtotal': imp_subtotal,
        }
        self.factura['detalles'].append(item)
        return True

    def AgregarDatoAdicional(self, t, c1, c2, c3, c4, c5, c6, **kwarg):
        "Agrego un tipo de dato adicional a una factura (interna)"
        op = {'t': t,
              'c1': c1, 'c2': c2, 'c3': c3, 'c4': c4, 'c5': c5, 'c6': c6, }
        self.factura['adicionales'].append(op)
        return True

    def AgregarFormaPago(self, codigo, tipo_tarjeta=None, numero_tarjeta=None,
                         swift_code=None, tipo_cuenta=None, numero_cuenta=None,
                         **kwarg):
        "Agrego una forma de pago a una factura (interna)"
        fp = {'codigo': codigo, 'tipo_tarjeta': tipo_tarjeta,
              'numero_tarjeta': numero_tarjeta, 'swift_code': swift_code,
              'tipo_cuenta': tipo_cuenta, 'numero_cuenta': numero_cuenta}
        self.factura['formas_pago'].append(fp)
        return True

    def EstablecerCampoItem(self, campo, valor):
        if self.factura['detalles'] and campo in self.factura['detalles'][-1]:
            self.factura['detalles'][-1][campo] = valor
            return True
        else:
            return False

    @inicializar_y_capturar_excepciones
    def AutorizarComprobante(self):
        f = self.factura
        # contruyo la estructura a convertir en XML:
        fact = {
            'codigoTipoDocumento': f['tipo_doc'], 'numeroDocumento': f['nro_doc'],
            'codigoTipoComprobante': f['tipo_cbte'], 'numeroPuntoVenta': f['punto_vta'],
            'codigoTipoAutorizacion': 'E',
            'numeroComprobante': f['cbte_nro'],
            'importeTotal': f['imp_total'], 'importeNoGravado': f['imp_tot_conc'],
            'idImpositivo': f['id_impositivo'],
            'codigoPais': f['cod_pais'], 'domicilioReceptor': f['domicilio'],
            'codigoRelacionEmisorReceptor': f['cod_relacion'],
            'importeGravado': f['imp_neto'],
            'importeSubtotal': f['imp_subtotal'],  # 'imp_iva': imp_iva,
            'importeOtrosTributos': f['tributos'] and f['imp_trib'] or None,
            'importeExento': f['imp_op_ex'], 'importeReintegro': f['imp_reintegro'],
            'fechaEmision': f['fecha_cbte'],
            'codigoMoneda': f['moneda_id'], 'cotizacionMoneda': f['moneda_ctz'],
            'observaciones': f['observaciones'],
            'fechaVencimientoPago': f.get('fecha_venc_pago'),
            'fechaServicioDesde': f.get('fecha_serv_desde'),
            'fechaServicioHasta': f.get('fecha_serv_hasta'),
            'arrayComprobantesAsociados': f['cbtes_asoc'] and [{'comprobanteAsociado': {
                'codigoTipoComprobante': cbte_asoc['tipo'],
                'numeroPuntoVenta': cbte_asoc['pto_vta'],
                'numeroComprobante': cbte_asoc['nro'],
            }} for cbte_asoc in f['cbtes_asoc']] or None,
            'arrayOtrosTributos': f['tributos'] and [{'otroTributo': {
                'codigo': tributo['tributo_id'],
                'descripcion': tributo['desc'],
                'baseImponible': tributo['base_imp'],
                'importe': tributo['importe'],
            }} for tributo in f['tributos']] or None,
            'arraySubtotalesIVA': f['iva'] and [{'subtotalIVA': {
                'codigo': iva['iva_id'],
                'importe': iva['importe'],
            }} for iva in f['iva']] or None,
            'arrayItems': f['detalles'] and [{'item': {
                'tipo': it['tipo'],
                'codigoTurismo': it['cod_tur'],
                'codigo': it['codigo'],
                'descripcion': it['ds'],
                'codigoAlicuotaIVA': it['iva_id'],
                'importeIVA': it['imp_iva'] if int(f['tipo_cbte']) not in (6, 7, 8) and it['imp_iva'] is not None else None,
                'importeItem': it['imp_subtotal'],
            }} for it in f['detalles']] or None,
            'arrayDatosAdicionales': [
                {'tipoDatoAdicional': ta} for ta in f['adicionales']] or None,
            'arrayFormasPago': [
                {'formaPago': {
                    'codigo': fp['codigo'],
                    'tipoTarjeta': fp['tipo_tarjeta'],
                    'numeroTarjeta': fp['numero_tarjeta'],
                    'swiftCode': fp['swift_code'],
                    'tipoCuenta': fp['tipo_cuenta'],
                    'numeroCuenta': fp['numero_cuenta'],
                }} for fp in f['formas_pago']] or None,
        }

        res = self.client.autorizarComprobante(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            comprobanteRequest=fact,
        )
        ret = res.get('autorizarComprobanteReturn', {})

        # Reprocesar en caso de error (recuperar CAE emitido anteriormente)
        if self.Reprocesar and ('arrayErrores' in ret):
            for error in ret['arrayErrores']:
                err_code = error['codigoDescripcion']['codigo']
                if ret['resultado'] == 'R' and err_code == 102:
                    # guardo los mensajes xml originales
                    xml_request = self.client.xml_request
                    xml_response = self.client.xml_response
                    cae = self.ConsultarComprobante(f['tipo_cbte'], f['punto_vta'], f['cbt_desde'], reproceso=True)
                    if cae and self.EmisionTipo == 'CAE':
                        self.Reproceso = 'S'
                        self.Resultado = 'A'  # verificar O
                        return cae
                    self.Reproceso = 'N'
                    # reestablesco los mensajes xml originales
                    self.client.xml_request = xml_request
                    self.client.xml_response = xml_response

        self.Resultado = ret.get('resultado', "")  # u'A'
        if self.Resultado in ("A", "O"):
            cbteresp = ret['comprobanteResponse']
            self.FechaCbte = cbteresp['fechaEmision'].strftime("%Y/%m/%d")
            self.CbteNro = cbteresp['numeroComprobante']  # 1L
            self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
            # self. = cbteresp['cuit'] # 20267565393L
            # self. = cbteresp['codigoTipoComprobante']
            self.Vencimiento = cbteresp['fechaVencimientoCAE'].strftime("%Y/%m/%d")
            self.CAE = str(cbteresp['CAE'])  # 60423794871430L
        self.__analizar_errores(ret)

        for error in ret.get('arrayObservaciones', []):
            self.Observaciones.append("%(codigo)s: %(descripcion)s" % (
                error['codigoDescripcion']))
        self.Obs = '\n'.join(self.Observaciones)
        self.EmisionTipo = 'CAE'

        if 'evento' in ret:
            self.Evento = '%(codigo)s: %(descripcion)s' % ret['evento']
        return self.CAE

    @inicializar_y_capturar_excepciones
    def CAESolicitar(self):
        try:
            cae = self.AutorizarComprobante() or ''
            self.Excepcion = "OK!"
        except BaseException:
            cae = "ERR"
        finally:
            return cae

    @inicializar_y_capturar_excepciones
    def ConsultarUltimoComprobanteAutorizado(self, tipo_cbte, punto_vta):
        res = self.client.consultarUltimoComprobanteAutorizado(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codigoTipoComprobante=tipo_cbte,
            numeroPuntoVenta=punto_vta,
        )
        ret = res.get('consultarUltimoComprobanteAutorizadoReturn', {})
        nro = ret.get('numeroComprobante')
        self.__analizar_errores(ret)
        self.CbteNro = nro
        return nro is not None and str(nro) or 0

    CompUltimoAutorizado = ConsultarUltimoComprobanteAutorizado

    @inicializar_y_capturar_excepciones
    def ConsultarComprobante(self, tipo_cbte, punto_vta, cbte_nro, reproceso=False):
        "Recuperar los datos completos de un comprobante ya autorizado"
        res = self.client.consultarComprobanteTipoPVentaNro(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codigoTipoComprobante=tipo_cbte,
            numeroPuntoVenta=punto_vta,
            numeroComprobante=cbte_nro,
        )
        ret = res.get('consultarComprobanteReturn', {})
        # diferencias si hay reproceso:
        difs = []
        # analizo el resultado:
        if 'comprobante' in ret:
            cbteresp = ret['comprobante']
            if reproceso:
                    # verifico los campos registrados coincidan con los enviados:
                f = self.factura
                verificaciones = {
                    'codigoTipoComprobante': f['tipo_cbte'],
                    'numeroPuntoVenta': f['punto_vta'],
                    'codigoTipoDocumento': f['tipo_doc'],
                    'numeroDocumento': f['nro_doc'],
                    'numeroComprobante': f['cbt_desde'],
                    'numeroComprobante': f['cbt_hasta'],
                    'fechaEmision': f['fecha_cbte'],
                    'idImpositivo': f['id_impositivo'],
                    'codigoPais': f['cod_pais'], 'domicilioReceptor': f['domicilio'],
                    'codigoRelacionEmisorReceptor': f['cod_relacion'],
                    'importeTotal': decimal.Decimal(str(f['imp_total'])),
                    'importeNoGravado': decimal.Decimal(str(f['imp_tot_conc'])),
                    'importeGravado': decimal.Decimal(str(f['imp_neto'])),
                    'importeExento': decimal.Decimal(str(f['imp_op_ex'])),
                    'importeOtrosTributos': f['tributos'] and decimal.Decimal(str(f['imp_trib'])) or None,
                    'importeSubtotal': f['imp_subtotal'],
                    'importeReintegro': f['imp_reintegro'],
                    'codigoMoneda': f['moneda_id'],
                    'cotizacionMoneda': str(decimal.Decimal(str(f['moneda_ctz']))),
                    'arrayItems': [
                        {'item': {
                            'tipo': it['tipo'],
                            'codigoTurismo': it['cod_tur'],
                            'codigo': it['codigo'],
                            'descripcion': it['ds'],
                            'codigoAlicuotaIVA': decimal.Decimal(str(it['iva_id'])),
                            'importeIVA': decimal.Decimal(str(it['imp_iva'])) if int(f['tipo_cbte']) not in (6, 7, 8) and it['imp_iva'] is not None else None,
                            'importeItem': decimal.Decimal(str(it['imp_subtotal'])),
                        }}
                        for it in f['detalles']],
                    'arrayComprobantesAsociados': [
                        {'comprobanteAsociado': {
                            'codigoTipoComprobante': cbte_asoc['tipo'],
                            'numeroPuntoVenta': cbte_asoc['pto_vta'],
                            'numeroComprobante': cbte_asoc['nro']}}
                        for cbte_asoc in f['cbtes_asoc']],
                    'arrayOtrosTributos': [
                        {'otroTributo': {
                            'codigo': tributo['tributo_id'],
                            'descripcion': tributo['desc'],
                            'baseImponible': decimal.Decimal(str(tributo['base_imp'])),
                            'importe': decimal.Decimal(str(tributo['importe'])),
                        }}
                        for tributo in f['tributos']],
                    'arraySubtotalesIVA': [
                        {'subtotalIVA': {
                            'codigo': iva['iva_id'],
                            'importe': decimal.Decimal(str(iva['importe'])),
                        }}
                        for iva in f['iva']],
                }
                verifica(verificaciones, cbteresp, difs)
                if difs:
                    print("Diferencias:", difs)
                    self.log("Diferencias: %s" % difs)
            self.FechaCbte = cbteresp['fechaEmision'].strftime("%Y/%m/%d")
            self.CbteNro = cbteresp['numeroComprobante']  # 1L
            self.PuntoVenta = cbteresp['numeroPuntoVenta']  # 4000
            self.Vencimiento = cbteresp['fechaVencimiento'].strftime("%Y/%m/%d")
            self.ImpTotal = str(cbteresp['importeTotal'])
            self.CAE = str(cbteresp['codigoAutorizacion'])  # 60423794871430L
            self.EmisionTipo = cbteresp['codigoTipoAutorizacion'] == 'A' and 'CAEA' or 'CAE'
        self.__analizar_errores(ret)
        if not difs:
            return self.CAE

    @inicializar_y_capturar_excepciones
    def ConsultarTiposComprobante(self):
        "Este método permite consultar los tipos de comprobantes habilitados en este WS"
        res = self.client.consultarTiposComprobantes(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarTiposComprobantesReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposComprobantes']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposDocumento(self):
        res = self.client.consultarTiposDocumento(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarTiposDocumentoReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposDocumento']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposIVA(self):
        "Este método permite consultar los tipos de IVA habilitados en este ws"
        res = self.client.consultarTiposIVA(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarTiposIVAReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcionString']
                for p in ret['arrayTiposIVA']]

    @inicializar_y_capturar_excepciones
    def ConsultarCondicionesIVA(self):
        "Este método permite consultar los tipos de comprobantes habilitados en este WS"
        res = self.client.consultarCondicionesIVA(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarCondicionesIVAReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcionString']
                for p in ret['arrayCondicionesIVA']]

    @inicializar_y_capturar_excepciones
    def ConsultarMonedas(self):
        "Este método permite consultar los tipos de comprobantes habilitados en este WS"
        res = self.client.consultarMonedas(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarMonedasReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcionString']
                for p in ret['arrayTiposMoneda']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposItem(self):
        "Este método permite consultar los tipos de comprobantes habilitados en este WS"
        res = self.client.consultarTiposItem(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarTiposItemReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayTiposItem']]

    @inicializar_y_capturar_excepciones
    def ConsultarCodigosItemTurismo(self):
        "Este método permite consultar los códigos de los ítems de Turismo"
        res = self.client.consultarCodigosItemTurismo(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarCodigosItemTurismoReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcion']
                for p in ret['arrayCodigosItem']]

    @inicializar_y_capturar_excepciones
    def ConsultarTiposTributo(self):
        "Este método permite consultar los tipos de comprobantes habilitados en este WS"
        res = self.client.consultarTiposTributo(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = res['consultarTiposTributoReturn']
        return ["%(codigo)s: %(descripcion)s" % p['codigoDescripcionString']
                for p in ret['arrayTiposTributo']]

    @inicializar_y_capturar_excepciones
    def ConsultarCotizacion(self, moneda_id):
        "Este método permite consultar los tipos de comprobantes habilitados en este WS"
        ret = self.client.consultarCotizacion(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
            codigoMoneda=moneda_id,
        )
        self.__analizar_errores(ret)
        if 'cotizacionMoneda' in ret:
            return str(ret['cotizacionMoneda'])

    @inicializar_y_capturar_excepciones
    def ConsultarPuntosVenta(self, fmt="%(numeroPuntoVenta)s: bloqueado=%(bloqueado)s baja=%(fechaBaja)s"):
        "Este método permite consultar los puntos de venta habilitados para CAE en este WS"
        res = self.client.consultarPuntosVenta(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit},
        )
        ret = []
        self.__analizar_errores(ret)
        for p in res['consultarPuntosVentaReturn'].get("arrayPuntosVenta", {}):
            p = p['puntoVenta']
            if 'fechaBaja' not in p:
                p['fechaBaja'] = ""
            ret.append(fmt % p if fmt else p)
        return ret

    @inicializar_y_capturar_excepciones
    def ConsultarPaises(self, sep="|"):
        "Recuperador de valores referenciales de códigos de Países"
        ret = self.client.consultarPaises(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit, })
        result = ret['consultarPaisesReturn']
        self.__analizar_errores(result)

        ret = []
        for u in result['arrayPaises']:
            u = u['codigoDescripcionString']
            try:
                r = {'codigo': u.get('codigo'), 'ds': u.get('descripcion'), }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [("\t%(codigo)s\t%(ds)s\t"
                     % it).replace("\t", sep) for it in ret]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def ConsultarCUITsPaises(self, sep="|"):
        "Recuperar lista de valores referenciales de CUIT de Países"
        ret = self.client.consultarCUITsPaises(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit, })
        result = ret['consultarCUITsPaisesReturn']
        self.__analizar_errores(result)

        ret = []
        for u in result['arrayCuitPaises']:
            u = u['codigoDescripcionString']
            try:
                r = {'codigo': u.get('codigo'), 'ds': u.get('descripcion'), }
            except Exception as e:
                print(e)

            ret.append(r)
        if sep:
            return [("\t%(codigo)s\t%(ds)s\t"
                     % it).replace("\t", sep) for it in ret]
        else:
            return ret

    @inicializar_y_capturar_excepciones
    def ConsultarTiposDatosAdicionales(self, sep="|"):
        "Recuperar lista de los datos adicionales a informar según RG."
        ret = self.client.consultarTiposDatosAdicionales(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit, })
        result = ret['consultarTiposDatosAdicionalesReturn']
        self.__analizar_errores(result)
        ret = []
        for u in result['arrayTiposDatosAdicionales']:
            u = u['codigoDescripcionString']
            r = {'codigo': u.get('codigo'), 'ds': u.get('descripcion'), }
            ret.append(r)
        return [("\t%(codigo)s\t%(ds)s\t"
                 % it).replace("\t", sep) for it in ret] if sep else ret

    @inicializar_y_capturar_excepciones
    def ConsultarFomasPago(self, sep="|"):
        "Recuperar lista de las formas de pago"
        ret = self.client.consultarFormasPago(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit, })
        result = ret['consultarFormasPagoReturn']
        self.__analizar_errores(result)
        ret = []
        for u in result['arrayFormasPago']:
            u = u['codigoDescripcion']
            r = {'codigo': u.get('codigo'), 'ds': u.get('descripcion'), }
            ret.append(r)
        return [("\t%(codigo)s\t%(ds)s\t"
                 % it).replace("\t", sep) for it in ret] if sep else ret

    @inicializar_y_capturar_excepciones
    def ConsultarTiposTarjeta(self, forma_pago=None, sep="|"):
        "Recuperar lista de los tipos de tarjeta habilitados"
        ret = self.client.consultarTiposTarjeta(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit, },
            formaPago=forma_pago)
        result = ret['consultarTiposTarjetaReturn']
        self.__analizar_errores(result)
        ret = []
        for u in result['arrayTiposTarjeta']:
            u = u['codigoDescripcion']
            r = {'codigo': u.get('codigo'), 'ds': u.get('descripcion'), }
            ret.append(r)
        return [("\t%(codigo)s\t%(ds)s\t"
                 % it).replace("\t", sep) for it in ret] if sep else ret

    @inicializar_y_capturar_excepciones
    def ConsultarTiposCuenta(self, sep="|"):
        "Recuperar lista de los tipos de tarjeta habilitados"
        ret = self.client.consultarTiposCuenta(
            authRequest={'token': self.Token, 'sign': self.Sign, 'cuitRepresentada': self.Cuit, })
        result = ret['consultarTiposCuentaReturn']
        self.__analizar_errores(result)
        ret = []
        for u in result['arrayTiposCuenta']:
            u = u['codigoDescripcion']
            r = {'codigo': u.get('codigo'), 'ds': u.get('descripcion'), }
            ret.append(r)
        return [("\t%(codigo)s\t%(ds)s\t"
                 % it).replace("\t", sep) for it in ret] if sep else ret


def main():
    "Función principal de pruebas (obtener CAE)"
    import os
    import time

    DEBUG = '--debug' in sys.argv

    # obteniendo el TA para pruebas
    from .wsaa import WSAA
    ta = WSAA().Autenticar("wsct", "reingart.crt", "reingart.key")

    wsct = WSCT()
    wsct.SetTicketAcceso(ta)
    wsct.Cuit = "20267565393"

    cache = ""
    if "--prod" in sys.argv:
        wsdl = "https://serviciosjava.afip.gob.ar/wsct/CTService?wsdl"
    else:
        wsdl = WSDL
    wsct.Conectar(cache, wsdl, cacert="conf/afip_ca_info.crt")

    if "--dummy" in sys.argv:
        print(wsct.client.help("dummy"))
        wsct.Dummy()
        print("AppServerStatus", wsct.AppServerStatus)
        print("DbServerStatus", wsct.DbServerStatus)
        print("AuthServerStatus", wsct.AuthServerStatus)

    if "--prueba" in sys.argv:
        # print wsct.client.help("autorizarComprobante").encode("latin1")
        try:
            tipo_cbte = 195
            punto_vta = 4000
            cbte_nro = wsct.ConsultarUltimoComprobanteAutorizado(tipo_cbte, punto_vta)
            fecha = datetime.datetime.now().strftime("%Y-%m-%d")
            tipo_doc = 80
            nro_doc = "50000000059"
            cbte_nro = int(cbte_nro) + 1
            id_impositivo = 9     # "Cliente del Exterior"
            cod_relacion = 3      # Alojamiento Directo a Turista No Residente
            imp_total = "101.00"
            imp_tot_conc = "0.00"
            imp_neto = "100.00"
            imp_trib = "1.00"
            imp_op_ex = "0.00"
            imp_subtotal = "100.00"
            imp_reintegro = -21.00      # validación AFIP 346
            cod_pais = 203
            domicilio = "Rua N.76 km 34.5 Alagoas"
            fecha_cbte = fecha
            moneda_id = 'PES'
            moneda_ctz = '1.000'
            obs = "Observaciones Comerciales, libre"

            wsct.CrearFactura(tipo_doc, nro_doc, tipo_cbte, punto_vta,
                              cbte_nro, imp_total, imp_tot_conc, imp_neto,
                              imp_subtotal, imp_trib, imp_op_ex, imp_reintegro,
                              fecha_cbte, id_impositivo, cod_pais, domicilio,
                              cod_relacion, moneda_id, moneda_ctz, obs)

            tributo_id = 99
            desc = 'Impuesto Municipal Matanza'
            base_imp = "100.00"
            alic = "1.00"
            importe = "1.00"
            wsct.AgregarTributo(tributo_id, desc, base_imp, alic, importe)

            iva_id = 5  # 21%
            base_imp = 100
            importe = 21
            wsct.AgregarIva(iva_id, base_imp, importe)

            tipo = 0    # Item General
            cod_tur = 1  # Servicio de hotelería - alojamiento sin desayuno
            codigo = "T0001"
            ds = "Descripcion del producto P0001"
            iva_id = 5
            imp_iva = 21.00
            imp_subtotal = 121.00
            wsct.AgregarItem(tipo, cod_tur, codigo, ds,
                             iva_id, imp_iva, imp_subtotal)

            codigo = 68             # tarjeta de crédito
            tipo_tarjeta = 99       # otra (ver tabla de parámetros)
            numero_tarjeta = "999999"
            swift_code = None
            tipo_cuenta = None
            numero_cuenta = None
            wsct.AgregarFormaPago(codigo, tipo_tarjeta, numero_tarjeta,
                                  swift_code, tipo_cuenta, numero_cuenta)

            print(wsct.factura)

            wsct.AutorizarComprobante()

            print("Resultado", wsct.Resultado)
            print("CAE", wsct.CAE)
            print("Vencimiento", wsct.Vencimiento)
            print("Reproceso", wsct.Reproceso)
            print("Errores", wsct.ErrMsg)

            print(wsct.Excepcion)
            print(wsct.ErrMsg)

            cae = wsct.CAE

            if cae:

                wsct.ConsultarComprobante(tipo_cbte, punto_vta, cbte_nro)
                print("CAE consulta", wsct.CAE, wsct.CAE == cae)
                print("NRO consulta", wsct.CbteNro, wsct.CbteNro == cbte_nro)
                print("TOTAL consulta", wsct.ImpTotal, wsct.ImpTotal == imp_total)

                wsct.AnalizarXml("XmlResponse")
                assert wsct.ObtenerTagXml('codigoAutorizacion') == str(wsct.CAE)
                assert wsct.ObtenerTagXml('codigoConcepto') == str(concepto)
                assert wsct.ObtenerTagXml('arrayItems', 0, 'item', 'unidadesMtx') == '123456'

        except BaseException:
            print(wsct.XmlRequest)
            print(wsct.XmlResponse)
            print(wsct.ErrCode)
            print(wsct.ErrMsg)

    if "--ptosventa" in sys.argv:
        print(wsct.ConsultarPuntosVenta())

    if "--parametros" in sys.argv:
        print(wsct.ConsultarTiposDatosAdicionales())
        print(wsct.ConsultarTiposComprobante())
        print(wsct.ConsultarTiposDocumento())
        print(wsct.ConsultarTiposIVA())
        print(wsct.ConsultarCondicionesIVA())
        print(wsct.ConsultarMonedas())
        print(wsct.ConsultarTiposItem())
        print(wsct.ConsultarCodigosItemTurismo())
        print(wsct.ConsultarTiposTributo())
        print(wsct.ConsultarFomasPago())
        for forma_pago in wsct.ConsultarFomasPago(sep=None)[:2]:
            print(wsct.ConsultarTiposTarjeta(forma_pago["codigo"]))
        print(wsct.ConsultarTiposCuenta())
        print("\n".join(wsct.ConsultarPaises()))
        print("\n".join(wsct.ConsultarCUITsPaises()))

    if "--cotizacion" in sys.argv:
        print(wsct.ConsultarCotizacionMoneda('DOL'))


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = WSCT.InstallDir = get_install_dir()


if __name__ == '__main__':

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSCT)
    else:
        main()



#!/usr/bin/python
# -*- coding: utf8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.


import shelve
import sys
import datetime
import decimal
import os
from .utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I, json, BaseWS, inicializar_y_capturar_excepciones, get_install_dir
from . import utils
from fpdf import Template
from pysimplesoap.client import SoapFault
import pprint
import traceback
"""Módulo para obtener código de autorización electrónica (CAE) para
Liquidación Única Mensual (lechería) del web service WSLUM de AFIP
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2016 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.03b"

LICENCIA = """
wslum.py: Interfaz para generar Código de Autorización Electrónica (CAE) para
          Liquidación Única Mensual de lechería (LumService)
Copyright (C) 2016 Mariano Reingart reingart@gmail.com
http://www.sistemasagiles.com.ar/trac/wiki/LiquidacionUnicaMensualLecheria

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo respetando la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

AYUDA = """
Opciones:
  --ayuda: este mensaje

  --debug: modo depuración (detalla y confirma las operaciones)
  --formato: muestra el formato de los archivos de entrada/salida
  --prueba: genera y autoriza una liquidación de prueba (no usar en producción!)
  --xml: almacena los requerimientos y respuestas XML (depuración)
  --json: utilizar formato json para el archivo de intercambio
  --dummy: consulta estado de servidores

  --autorizar: Autorizar Liquidación Única Mensual (lechería) (generarLiquidacion)
  --ult: Consulta el último número de orden registrado en AFIP
         (consultarUltimoComprobanteXPuntoVenta)
  --consultar: Consulta una liquidación registrada en AFIP
         (consultarLiquidacionXNroComprobante / consultarLiquidacionXCAE)

  --pdf: descarga la liquidación en formato PDF
  --mostrar: muestra el documento PDF generado (usar con --pdf)
  --imprimir: imprime el documento PDF generado (usar con --mostrar y --pdf)

  --provincias: obtiene el listado de provincias (código/descripción)
  --localidades: obtiene el listado de localidades para una provincia
  --bonificaciones_penalizaciones: obtiene el listado de tributos
  --otros_impuestos: obtiene el listado de las retenciones de tabaco
  --puntosventa: obtiene el listado de puntos de venta habilitados

Ver wslum.ini para parámetros de configuración (URL, certificados, etc.)"
"""


# importo funciones compartidas:


WSDL = "https://fwshomo.afip.gov.ar/wslum/LumService?wsdl"
#WSDL = "https://serviciosjava.afip.gov.ar/wslum/LumService?wsdl"

DEBUG = False
XML = False
CONFIG_FILE = "wslum.ini"
HOMO = True


class WSLUM(BaseWS):
    "Interfaz para el WebService de Liquidación Única Mensual (lechería)"
    _public_methods_ = ['Conectar', 'Dummy', 'SetTicketAcceso', 'DebugLog',
                        'AutorizarLiquidacion',
                        'CrearLiquidacion',
                        'AgregarTambero', 'AgregarCondicionVenta',
                        'AgregarTambo', 'AgregarUbicacionTambo',
                        'AgregarBalanceLitrosPorcentajesSolidos',
                        'AgregarConceptosBasicosMercadoInterno',
                        'AgregarConceptosBasicosMercadoExterno',
                        'AgregarBonificacionPenalizacion',
                        'AgregarOtroImpuesto',
                        'AgregarRemito',
                        'ConsultarLiquidacion', 'ConsultarUltimoComprobante',
                        'AgregarAjuste',
                        'LeerDatosLiquidacion',
                        'ConsultarBonificacionesPenalizaciones',
                        'ConsultarOtrosImpuestos',
                        'ConsultarPuntosVentas',
                        'ConsultarProvincias', 'ConsultarLocalidades',
                        'MostrarPDF',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        'SetParametros', 'SetParametro', 'GetParametro',
                        ]
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
                      'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
                      'Excepcion', 'ErrCode', 'ErrMsg', 'LanzarExcepciones', 'Errores',
                      'XmlRequest', 'XmlResponse', 'Version', 'Traceback', 'InstallDir',
                      'CAE', 'NroComprobante', 'FechaComprobante',
                      'AlicuotaIVA', 'TotalNeto', 'ImporteIVA',
                      'TotalBonificacionesCalidad', 'TotalPenalizacionesCalidad',
                      'TotalBonificacionesComerciales', 'TotalDebitosComerciales',
                      'TotalOtrosImpuestos', 'Total',
                      ]
    _reg_progid_ = "WSLUM"
    _reg_clsid_ = "{4CBB2DF8-7AAE-434E-916D-9D663BB1CAFC}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    LanzarExcepciones = False
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.errores = []
        self.CAE = ""
        self.NroComprobante = self.FechaComprobante = ''
        self.AlicuotaIVA = self.TotalNeto = self.ImporteIVA = None
        self.TotalBonificacionesCalidad = None
        self.TotalPenalizacionesCalidad = None
        self.TotalBonificacionesComerciales = None
        self.TotalDebitosComerciales = None
        self.TotalOtrosImpuestos = None
        self.Total = None
        self.datos = {}

    @inicializar_y_capturar_excepciones
    def Conectar(self, cache=None, url="", proxy="", wrapper="", cacert=None, timeout=30):
        "Establecer la conexión a los servidores de la AFIP"
        # llamo al constructor heredado:
        ok = BaseWS.Conectar(self, cache, url, proxy, wrapper, cacert, timeout)
        if False and ok:
            # corrijo ubicación del servidor (puerto htttp 80 en el WSDL)
            location = self.client.services['LumService']['ports']['LumEndPoint']['location']
            if location.startswith("http://"):
                print("Corrigiendo WSDL ...", location, end=' ')
                location = location.replace("http://", "https://").replace(":80", ":443")
                self.client.services['LumService']['ports']['LumEndPoint']['location'] = location
                print(location)
        return ok

    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        errores = []
        if 'errores' in ret:
            errores.extend(ret['errores'])
        if errores:
            self.Errores = ["%(codigo)s: %(descripcion)s" % err['error'][0]
                            for err in errores]
            self.errores = [
                {'codigo': err['error'][0]['codigo'],
                 'descripcion': err['error'][0]['descripcion'].replace("\n", "")
                 .replace("\r", "")}
                for err in errores]
            self.ErrCode = ' '.join(self.Errores)
            self.ErrMsg = '\n'.join(self.Errores)

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.dummy()['respuesta']
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])
        return True

    @inicializar_y_capturar_excepciones
    def CrearLiquidacion(self, tipo_cbte, pto_vta, nro_cbte, fecha, periodo,
                         iibb_adquirente=None, domicilio_sede=None,
                         inscripcion_registro_publico=None,
                         datos_adicionales=None, alicuota_iva=None, **kwargs):
        "Inicializa internamente los datos de una liquidación para autorizar"
        # creo el diccionario con los campos generales de la liquidación:
        liq = {'tipoComprobante': tipo_cbte, 'puntoVenta': pto_vta,
               'nroComprobante': nro_cbte, 'fechaComprobante': fecha,
               'periodo': periodo, 'iibbAdquirente': iibb_adquirente,
               'domicilioSede': domicilio_sede,
               'inscripcionRegistroPublico': inscripcion_registro_publico,
               'datosAdicionales': datos_adicionales,
               'alicuotaIVA': alicuota_iva,
               }
        liq["condicionVenta"] = []
        self.solicitud = dict(liquidacion=liq,
                              bonificacionPenalizacion=[],
                              otroImpuesto=[],
                              remito=[]
                              )
        return True

    @inicializar_y_capturar_excepciones
    def AgregarCondicionVenta(self, codigo, descripcion=None, **kwargs):
        "Agrego una o más condicion de venta a la liq."
        cond = {'codigo': codigo, 'descripcion': descripcion}
        self.solicitud['liquidacion']['condicionVenta'].append(cond)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarTambero(self, cuit, iibb=None, **kwargs):
        "Agrego los datos del productor a la liq."
        tambero = {'cuit': cuit, 'iibb': iibb}
        self.solicitud['tambero'] = tambero
        return True

    @inicializar_y_capturar_excepciones
    def AgregarTambo(self, nro_tambo_interno, nro_renspa,
                     fecha_venc_cert_tuberculosis, fecha_venc_cert_brucelosis,
                     nro_tambo_provincial=None, **kwargs):
        "Agrego los datos del productor a la liq."
        tambo = {'nroTamboInterno': nro_tambo_interno,
                 'nroTamboProvincial': nro_tambo_provincial,
                 'nroRenspa': nro_renspa,
                 'ubicacionTambo': {},
                 'fechaVencCertTuberculosis': fecha_venc_cert_tuberculosis,
                 'fechaVencCertBrucelosis': fecha_venc_cert_brucelosis}
        self.solicitud['tambo'] = tambo
        return True

    @inicializar_y_capturar_excepciones
    def AgregarUbicacionTambo(self, latitud, longitud, domicilio,
                              cod_localidad, cod_provincia, codigo_postal,
                              nombre_partido_depto, **kwargs):
        "Agrego los datos del productor a la liq."
        ubic_tambo = {'latitud': latitud,
                      'longitud': longitud,
                      'domicilio': domicilio,
                      'codLocalidad': cod_localidad,
                      'codProvincia': cod_provincia,
                      'nombrePartidoDepto': nombre_partido_depto,
                      'codigoPostal': codigo_postal}
        self.solicitud['tambo']['ubicacionTambo'] = ubic_tambo
        return True

    @inicializar_y_capturar_excepciones
    def AgregarBalanceLitrosPorcentajesSolidos(self, litros_remitidos, litros_decomisados,
                                               kg_grasa, kg_proteina, **kwargs):
        "Agrega balance litros y porcentajes sólidos a la liq. (obligatorio)"
        d = {'litrosRemitidos': litros_remitidos,
             'litrosDecomisados': litros_decomisados,
             'kgGrasa': kg_grasa,
             'kgProteina': kg_proteina}
        self.solicitud['balanceLitrosPorcentajesSolidos'] = d

    @inicializar_y_capturar_excepciones
    def AgregarConceptosBasicosMercadoInterno(self, kg_produccion_gb, precio_por_kg_produccion_gb,
                                              kg_produccion_pr, precio_por_kg_produccion_pr,
                                              kg_crecimiento_gb, precio_por_kg_crecimiento_gb,
                                              kg_crecimiento_pr, precio_por_kg_crecimiento_pr,
                                              **kwargs):
        "Agrega balance litros y porcentajes sólidos (mercado interno)"
        d = {'kgProduccionGB': kg_produccion_gb,
             'precioPorKgProduccionGB': precio_por_kg_produccion_gb,
             'kgProduccionPR': kg_produccion_pr,
             'precioPorKgProduccionPR': precio_por_kg_produccion_pr,
             'kgCrecimientoGB': kg_crecimiento_gb,
             'precioPorKgCrecimientoGB': precio_por_kg_crecimiento_gb,
             'kgCrecimientoPR': kg_crecimiento_pr,
             'precioPorKgCrecimientoPR': precio_por_kg_crecimiento_pr}
        self.solicitud['conceptosBasicosMercadoInterno'] = d
        return True

    @inicializar_y_capturar_excepciones
    def AgregarConceptosBasicosMercadoExterno(self, kg_produccion_gb, precio_por_kg_produccion_gb,
                                              kg_produccion_pr, precio_por_kg_produccion_pr,
                                              kg_crecimiento_gb, precio_por_kg_crecimiento_gb,
                                              kg_crecimiento_pr, precio_por_kg_crecimiento_pr,
                                              **kwargs):
        "Agrega balance litros y porcentajes sólidos (mercado externo)"
        d = {'kgProduccionGB': kg_produccion_gb,
             'precioPorKgProduccionGB': precio_por_kg_produccion_gb,
             'kgProduccionPR': kg_produccion_pr,
             'precioPorKgProduccionPR': precio_por_kg_produccion_pr,
             'kgCrecimientoGB': kg_crecimiento_gb,
             'precioPorKgCrecimientoGB': precio_por_kg_crecimiento_gb,
             'kgCrecimientoPR': kg_crecimiento_pr,
             'precioPorKgCrecimientoPR': precio_por_kg_crecimiento_pr}
        self.solicitud['conceptosBasicosMercadoExterno'] = d

    @inicializar_y_capturar_excepciones
    def AgregarBonificacionPenalizacion(self, codigo, detalle, resultado=None,
                                        porcentaje=None, importe=None, **kwargs):
        "Agrega la información referente a las bonificaciones o penalizaciones"
        ret = dict(codBonificacionPenalizacion=codigo, detalle=detalle,
                   resultado=resultado, porcentajeAAplicar=porcentaje,
                   importe=importe)
        self.solicitud['bonificacionPenalizacion'].append(ret)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarOtroImpuesto(self, tipo, base_imponible, alicuota, detalle=None):
        "Agrega la información referente a otros tributos de la liquidación"
        trib = dict(tipo=tipo, baseImponible=base_imponible, alicuota=alicuota,
                    detalle=detalle)
        self.solicitud['otroImpuesto'].append(trib)
        return True

    @inicializar_y_capturar_excepciones
    def AgregarRemito(self, nro_remito):
        "Agrega la información referente a los remitos (multiples)"
        self.solicitud['remito'].append(nro_remito)
        return True

    @inicializar_y_capturar_excepciones
    def AutorizarLiquidacion(self):
        "Generar o ajustar una liquidación única y obtener del CAE"
        # limpio los elementos que no correspondan por estar vacios:
        for campo in ["bonificacionPenalizacion", "otroImpuesto"]:
            if campo in self.solicitud and not self.solicitud[campo]:
                del self.solicitud[campo]
        # llamo al webservice:
        ret = self.client.generarLiquidacion(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud=self.solicitud,
        )
        # analizo la respusta
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        liqs = ret.get('liquidacion', [])
        liq = liqs[0] if liqs else None
        self.AnalizarLiquidacion(liq)
        return True

    def AnalizarLiquidacion(self, liq):
        "Método interno para analizar la respuesta de AFIP"
        # proceso los datos básicos de la liquidación (devuelto por consultar):
        if liq:
            cab = liq['encabezado']
            self.CAE = str(cab['cae'])
            self.FechaComprobante = str(cab['fechaComprobante'])
            self.NroComprobante = cab['nroComprobante']
            tot = liq['resumenTotales']
            self.AlicuotaIVA = tot['alicuotaIVA']
            self.TotalNeto = tot['totalNetoLiquidacion']
            self.ImporteIVA = tot['importeIVA']
            self.TotalBonificacionesCalidad = tot['totalBonificacionesCalidad']
            self.TotalPenalizacionesCalidad = tot['totalPenalizacionesCalidad']
            self.TotalBonificacionesComerciales = tot['totalBonificacionesComerciales']
            self.TotalDebitosComerciales = tot['totalDebitosComerciales']
            self.TotalOtrosImpuestos = tot['totalOtrosImpuestos']
            self.Total = tot['totalLiquidacion']

            # parámetros de salida:
            self.params_out = dict(
                tipo_cbte=liq['encabezado']['tipoComprobante'],
                pto_vta=liq['encabezado']['puntoVenta'],
                nro_cbte=liq['encabezado']['nroComprobante'],
                fecha=str(liq['encabezado']['fechaComprobante']),
                cae=str(liq['encabezado']['cae']),
                domicilio_comprador=liq['encabezado']['domicilioComprador'],
                tambero=dict(
                    cuit=liq['tambero']['cuit'],
                    iibb=liq['tambero']['iibb'],
                    razon_social=liq['tambero']['razonSocial'],
                    situacion_iva=liq['tambero']['situacionIVA'],
                    domicilio_fiscal=liq['tambero']['domicilioFiscal'],
                    provincia=liq['tambero']['provincia'],
                    cod_postal=liq['tambero']['codPostal'],
                ),
                resumen_kg_remitidos=liq['resumenKgRemitidos'],
                alicuota_iva=liq['resumenTotales']['alicuotaIVA'],
                importe_iva=liq['resumenTotales']['importeIVA'],
                total_neto=liq['resumenTotales']['totalNetoLiquidacion'],
                total_bonificaciones_calidad=liq['resumenTotales']['totalBonificacionesCalidad'],
                total_penalizaciones_calidad=liq['resumenTotales']['totalPenalizacionesCalidad'],
                total_bonificaciones_comerciales=liq['resumenTotales']['totalBonificacionesComerciales'],
                total_debitos_comerciales=liq['resumenTotales']['totalDebitosComerciales'],
                total_otros_impuestos=liq['resumenTotales']['totalOtrosImpuestos'],
                total=liq['resumenTotales']['totalLiquidacion'],
                bonificacion_penalizacion=[],
                remitos=[],
                otro_impuesto=[],
                pdf=liq.get('pdf'),
            )
            for ret in liq.get('bonificacionPenalizacion', []):
                self.params_out['bonificacion_penalizacion'].append(dict(
                    retencion_codigo=ret['codigo'],
                    retencion_importe=ret['importe'],
                ))
            for trib in liq.get('otroImpuesto', []):
                self.params_out['otro_impuesto'].append(dict(
                    tributo_descripcion=trib.get('descripcion', ""),
                    tributo_base_imponible=trib['baseImponible'],
                    tributo_alicuota=trib['alicuota'],
                    tributo_codigo=trib['codigo'],
                    tributo_importe=trib['importe'],
                ))
            if DEBUG:
                import pprint
                pprint.pprint(self.params_out)
        self.params_out['errores'] = self.errores

    @inicializar_y_capturar_excepciones
    def AgregarAjuste(self, cai, tipo_cbte, pto_vta, nro_cbte, cae_a_ajustar):
        "Agrega comprobante a ajustar"
        ajuste = self.solicitud['liquidacion']['ajuste'] = {}
        if cae_a_ajustar:
            ajuste['caeAAjustar'] = cae_a_ajustar
        if cai:
            cbte = dict(cai=cai, caeAAjustar=cae_a_ajustar,
                        tipoComprobante=tipo_cbte, puntoVenta=pto_vta,
                        nroComprobante=nro_cbte)
            ajuste['formularioPapel'] = cbte
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarLiquidacion(self, tipo_cbte=None, pto_vta=None, nro_cbte=None,
                             cae=None, cuit_comprador=None, pdf="liq.pdf"):
        "Consulta una liquidación por No de Comprobante o CAE"
        if cae:
            ret = self.client.consultarLiquidacionPorCae(
                auth={
                    'token': self.Token, 'sign': self.Sign,
                    'cuit': self.Cuit, },
                solicitud={
                    'cae': cae,
                    'pdf': pdf and True or False,
                },
            )
        else:
            ret = self.client.consultarLiquidacionPorNroComprobante(
                auth={
                    'token': self.Token, 'sign': self.Sign,
                    'cuit': self.Cuit, },
                solicitud={
                    'cuitComprador': cuit_comprador,
                    'puntoVenta': pto_vta,
                    'nroComprobante': nro_cbte,
                    'tipoComprobante': tipo_cbte,
                    'pdf': pdf and True or False,
                },
            )
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        if 'liquidacion' in ret:
            liqs = ret.get('liquidacion', [])
            liq = liqs[0] if liqs else None
            self.AnalizarLiquidacion(liq)
            # guardo el PDF si se indico archivo y vino en la respuesta:
            if pdf and 'pdf' in liq:
                open(pdf, "wb").write(liq['pdf'])
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarUltimoComprobante(self, tipo_cbte=151, pto_vta=1):
        "Consulta el último No de Comprobante registrado"
        ret = self.client.consultarUltimoNroComprobantePorPtoVta(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud={
                'puntoVenta': pto_vta,
                'tipoComprobante': tipo_cbte},
        )
        ret = ret['respuesta']
        self.__analizar_errores(ret)
        self.NroComprobante = ret['nroComprobante']
        return True

    def ConsultarProvincias(self, sep="||"):
        "Consulta las provincias habilitadas"
        ret = self.client.consultarProvincias(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('provincia', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarLocalidades(self, cod_provincia, sep="||"):
        "Consulta las localidades habilitadas"
        ret = self.client.consultarLocalidadesPorProvincia(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
            solicitud={'codProvincia': cod_provincia},
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('localidad', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarCondicionesVenta(self, sep="||"):
        "Retorna un listado de códigos y descripciones de las condiciones de ventas"
        ret = self.client.consultarCondicionesVenta(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('condicionVenta', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarOtrosImpuestos(self, sep="||"):
        "Retorna un listado de tributos con código, descripción y signo."
        ret = self.client.consultarOtrosImpuestos(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('otroImpuesto', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def ConsultarBonificacionesPenalizaciones(self, sep="||"):
        "Retorna un listado de bonificaciones/penalizaciones con código y descripción"
        ret = self.client.consultarBonificacionesPenalizaciones(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        self.XmlResponse = self.client.xml_response
        array = ret.get('tipo', [])
        if sep is None:
            # sin separador, devuelve un diccionario con clave cod_variadedad
            # y valor: {"descripcion": ds_variedad, "clases": lista_clases}
            # siendo lista_clases = [{'codigo': ..., 'descripcion': ...}]
            return dict([(it['codigo'], {'descripcion': it['descripcion'],
                                         'subtipo': it['subtipo']})
                         for it in array])
        else:
            # con separador, devuelve una lista de strings:
            # || cod.variedad || desc.variedad || desc.clase || cod.clase ||
            ret = []
            for it in array:
                for subtipo in it['subtipo']:
                    ret.append(
                        ("%s %%s %s %%s %s %%s %s %%s %s %%s %s %%s %s" %
                            (sep, sep, sep, sep, sep, sep, sep)) %
                        (it['codigo'], it['descripcion'],
                         subtipo['descripcion'], subtipo['codigo'],
                         subtipo['valor'], subtipo['signo'])
                    )
            return ret

    def ConsultarPuntosVentas(self, sep="||"):
        "Retorna los puntos de ventas autorizados para la utilizacion de WS"
        ret = self.client.consultarPuntosVenta(
            auth={
                'token': self.Token, 'sign': self.Sign,
                'cuit': self.Cuit, },
        )['respuesta']
        self.__analizar_errores(ret)
        array = ret.get('puntoVenta', [])
        if sep is None:
            return dict([(it['codigo'], it['descripcion']) for it in array])
        else:
            return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['codigo'], it['descripcion']) for it in array]

    def MostrarPDF(self, archivo, imprimir=False):
        try:
            if sys.platform == "linux2":
                os.system("evince ""%s""" % archivo)
            else:
                operation = imprimir and "print" or ""
                os.startfile(archivo, operation)
            return True
        except Exception as e:
            self.Excepcion = str(e)
            return False


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = WSLUM.InstallDir = get_install_dir()


if __name__ == '__main__':
    if '--ayuda' in sys.argv:
        print(LICENCIA)
        print(AYUDA)
        sys.exit(0)
    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato in []:
            comienzo = 1
            print("=== %s ===" % msg)
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                dec = len(fmt) > 3 and fmt[3] or (tipo == 'I' and '2' or '')
                print(" * Campo: %-20s Posición: %3d Longitud: %4d Tipo: %s Decimales: %s" % (
                    clave, comienzo, longitud, tipo, dec))
                comienzo += longitud
        sys.exit(0)

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSLUM)
        sys.exit(0)

    import csv
    from configparser import SafeConfigParser

    from .wsaa import WSAA

    try:

        if "--version" in sys.argv:
            print("Versión: ", __version__)

        if len(sys.argv) > 1 and sys.argv[1].endswith(".ini"):
            CONFIG_FILE = sys.argv[1]
            print("Usando configuracion:", CONFIG_FILE)

        config = SafeConfigParser()
        config.read(CONFIG_FILE)
        CERT = config.get('WSAA', 'CERT')
        PRIVATEKEY = config.get('WSAA', 'PRIVATEKEY')
        CUIT = config.get('WSLUM', 'CUIT')
        ENTRADA = config.get('WSLUM', 'ENTRADA')
        SALIDA = config.get('WSLUM', 'SALIDA')

        if config.has_option('WSAA', 'URL') and not HOMO:
            WSAA_URL = config.get('WSAA', 'URL')
        else:
            WSAA_URL = None  # wsaa.WSAAURL
        if config.has_option('WSLUM', 'URL') and not HOMO:
            WSLUM_URL = config.get('WSLUM', 'URL')
        else:
            WSLUM_URL = WSDL

        PROXY = config.has_option('WSAA', 'PROXY') and config.get('WSAA', 'PROXY') or None
        CACERT = config.has_option('WSAA', 'CACERT') and config.get('WSAA', 'CACERT') or None
        WRAPPER = config.has_option('WSAA', 'WRAPPER') and config.get('WSAA', 'WRAPPER') or None

        if config.has_section('DBF'):
            conf_dbf = dict(config.items('DBF'))
            if DEBUG:
                print("conf_dbf", conf_dbf)
        else:
            conf_dbf = {}

        DEBUG = '--debug' in sys.argv
        XML = '--xml' in sys.argv

        if DEBUG:
            print("Usando Configuración:")
            print("WSAA_URL:", WSAA_URL)
            print("WSLUM_URL:", WSLUM_URL)
            print("CACERT", CACERT)
            print("WRAPPER", WRAPPER)
        # obteniendo el TA
        from .wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wslum", CERT, PRIVATEKEY, wsdl=WSAA_URL,
                             proxy=PROXY, wrapper=WRAPPER, cacert=CACERT)
        if not ta:
            pass  # sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)

        # cliente soap del web service
        wslum = WSLUM()
        wslum.LanzarExcepciones = True
        wslum.Conectar(url=WSLUM_URL, proxy=PROXY, wrapper=WRAPPER, cacert=CACERT)
        wslum.SetTicketAcceso(ta)
        wslum.Cuit = CUIT

        if '--dummy' in sys.argv:
            ret = wslum.Dummy()
            print("AppServerStatus", wslum.AppServerStatus)
            print("DbServerStatus", wslum.DbServerStatus)
            print("AuthServerStatus", wslum.AuthServerStatus)
            # sys.exit(0)

        if '--autorizar' in sys.argv:

            if '--prueba' in sys.argv:

                # Solicitud 1: Alta de liquidación
                wslum.CrearLiquidacion(tipo_cbte=27, pto_vta=1, nro_cbte=1,
                                       fecha="2015-12-31", periodo="2015/12",
                                       iibb_adquirente="123456789012345",
                                       domicilio_sede="Domicilio Administrativo",
                                       inscripcion_registro_publico="Nro IGJ",
                                       datos_adicionales="Datos Adicionales Varios",
                                       alicuota_iva=21.00)
                wslum.AgregarCondicionVenta(codigo=1, descripcion=None)
                if '--ajuste' in sys.argv:
                    wslum.AgregarAjuste(cai="10000000000000",
                                        tipo_cbte=0, pto_vta=0, nro_cbte=0,
                                        cae_a_ajustar="75521002437246")

                wslum.AgregarTambero(cuit=11111111111, iibb="123456789012345")

                wslum.AgregarTambo(nro_tambo_interno=123456789,
                                   nro_renspa="12.345.6.78901/12",
                                   fecha_venc_cert_tuberculosis="2015-01-01",
                                   fecha_venc_cert_brucelosis="2015-01-01",
                                   nro_tambo_provincial=100000000)
                wslum.AgregarUbicacionTambo(
                    latitud=-34.62987, longitud=-58.65155,
                    domicilio="Domicilio Tambo",
                    cod_localidad=10109, cod_provincia=1,
                    codigo_postal=1234,
                    nombre_partido_depto='Partido Tambo')

                wslum.AgregarBalanceLitrosPorcentajesSolidos(
                    litros_remitidos=11000, litros_decomisados=1000,
                    kg_grasa=100.00, kg_proteina=100.00)

                wslum.AgregarConceptosBasicosMercadoInterno(
                    kg_produccion_gb=100, precio_por_kg_produccion_gb=5.00,
                    kg_produccion_pr=100, precio_por_kg_produccion_pr=5.00,
                    kg_crecimiento_gb=0, precio_por_kg_crecimiento_gb=0.00,
                    kg_crecimiento_pr=0, precio_por_kg_crecimiento_pr=0.00)

                wslum.AgregarConceptosBasicosMercadoExterno(
                    kg_produccion_gb=0, precio_por_kg_produccion_gb=0.00,
                    kg_produccion_pr=0, precio_por_kg_produccion_pr=0.00,
                    kg_crecimiento_gb=0, precio_por_kg_crecimiento_gb=0.00,
                    kg_crecimiento_pr=0, precio_por_kg_crecimiento_pr=0.00)

                wslum.AgregarBonificacionPenalizacion(codigo=1,
                                                      detalle="opcional", resultado="400", porcentaje=10.00)
                wslum.AgregarBonificacionPenalizacion(codigo=10,
                                                      detalle="opcional", resultado="2.5", porcentaje=10.00)
                wslum.AgregarBonificacionPenalizacion(codigo=4,
                                                      detalle="opcional", resultado="400", porcentaje=10.00)
                wslum.AgregarBonificacionPenalizacion(codigo=5,
                                                      detalle="opcional", resultado="En Saneamiento",
                                                      porcentaje=10.00)

                wslum.AgregarOtroImpuesto(tipo=1, base_imponible=100.00,
                                          alicuota=10.00, detalle="")
                wslum.AgregarOtroImpuesto(tipo=9, base_imponible=100.00,
                                          alicuota=10.00,
                                          detalle="Detalle Otras Percepciones")
                wslum.AgregarOtroImpuesto(tipo=8, base_imponible=100.00,
                                          alicuota=10.00, detalle="")

                wslum.AgregarRemito(nro_remito="123456789012")
                wslum.AgregarRemito(nro_remito="123456789")

            else:
                # cargar un archivo de texto:
                with open("wslum.json", "r") as f:
                    wslum.solicitud = json.load(f, encoding="utf-8")

            if '--testing' in sys.argv:
                # mensaje de prueba (no realiza llamada remota),
                # usar solo si no está operativo, cargo respuesta:
                wslum.LoadTestXML("tests/xml/wslum_liq_test_pdf_response.xml")
                import json
                with open("wslum.json", "w") as f:
                    json.dump(wslum.solicitud, f, sort_keys=True, indent=4, encoding="utf-8",)

            print("Liquidacion: pto_vta=%s nro_cbte=%s tipo_cbte=%s" % (
                wslum.solicitud['liquidacion']['puntoVenta'],
                wslum.solicitud['liquidacion']['nroComprobante'],
                wslum.solicitud['liquidacion']['tipoComprobante'],
            ))

            if not '--dummy' in sys.argv:
                print("Autorizando...")
                ret = wslum.AutorizarLiquidacion()

            if wslum.Excepcion:
                print("EXCEPCION:", wslum.Excepcion, file=sys.stderr)
                if DEBUG:
                    print(wslum.Traceback, file=sys.stderr)
            print("Errores:", wslum.Errores)
            print("CAE", wslum.CAE)
            print("FechaComprobante", wslum.FechaComprobante)
            print("NroComprobante", wslum.NroComprobante)
            print("TotalNeto", wslum.TotalNeto)
            print("AlicuotaIVA", wslum.AlicuotaIVA)
            print("ImporteIVA", wslum.ImporteIVA)
            print("TotalBonificacionesCalidad", wslum.TotalBonificacionesCalidad)
            print("TotalPenalizacionesCalidad", wslum.TotalPenalizacionesCalidad)
            print("TotalBonificacionesComerciales", wslum.TotalBonificacionesComerciales)
            print("TotalDebitosComerciales", wslum.TotalDebitosComerciales)
            print("TotalOtrosImpuestos", wslum.TotalOtrosImpuestos)
            print("Total", wslum.Total)

            pdf = wslum.GetParametro("pdf")
            if pdf:
                open("liq.pdf", "wb").write(pdf)

            if '--testing' in sys.argv:
                assert wslum.CAE == "75521002437246"

            if DEBUG:
                pprint.pprint(wslum.params_out)

            if "--guardar" in sys.argv:
                # grabar un archivo de texto (intercambio) con el resultado:
                liq = wslum.params_out.copy()
                if "pdf" in liq:
                    del liq["pdf"]                  # eliminador binario
                with open("wslum_salida.json", "w") as f:
                    json.dump(liq, f, default=str,
                              indent=2, sort_keys=True, encoding="utf-8")

        if '--consultar' in sys.argv:
            tipo_cbte = 27
            pto_vta = 1
            nro_cbte = 0
            cuit = None
            try:
                tipo_cbte = sys.argv[sys.argv.index("--consultar") + 1]
                pto_vta = sys.argv[sys.argv.index("--consultar") + 2]
                nro_cbte = sys.argv[sys.argv.index("--consultar") + 3]
                cuit = sys.argv[sys.argv.index("--consultar") + 4]
            except IndexError:
                pass
            if '--testing' in sys.argv:
                # mensaje de prueba (no realiza llamada remota),
                # usar solo si no está operativo, cargo prueba:
                wslum.LoadTestXML("tests/xml/wslum_cons_test.xml")
            print("Consultando: tipo_cbte=%s pto_vta=%s nro_cbte=%s" % (tipo_cbte, pto_vta, nro_cbte))
            ret = wslum.ConsultarLiquidacion(tipo_cbte, pto_vta, nro_cbte,
                                             cuit_comprador=cuit)
            print("CAE", wslum.CAE)
            print("Errores:", wslum.Errores)

            if DEBUG:
                pprint.pprint(wslum.params_out)

            if '--mostrar' in sys.argv and pdf:
                wslum.MostrarPDF(archivo=pdf,
                                 imprimir='--imprimir' in sys.argv)

        if '--ult' in sys.argv:
            tipo_cbte = 27
            pto_vta = 1
            try:
                tipo_cbte = sys.argv[sys.argv.index("--ult") + 1]
                pto_vta = sys.argv[sys.argv.index("--ult") + 2]
            except IndexError:
                pass

            print("Consultando ultimo nro_cbte para pto_vta=%s" % pto_vta, end=' ')
            ret = wslum.ConsultarUltimoComprobante(tipo_cbte, pto_vta)
            if wslum.Excepcion:
                print("EXCEPCION:", wslum.Excepcion, file=sys.stderr)
                if DEBUG:
                    print(wslum.Traceback, file=sys.stderr)
            print("Ultimo Nro de Comprobante", wslum.NroComprobante)
            print("Errores:", wslum.Errores)
            sys.exit(0)

        # Recuperar parámetros:

        if '--provincias' in sys.argv:
            ret = wslum.ConsultarProvincias()
            print("\n".join(ret))

        if '--localidades' in sys.argv:
            try:
                cod_provincia = sys.argv[sys.argv.index("--localidades") + 1]
            except BaseException:
                cod_provincia = input("Codigo Provincia:")
            ret = wslum.ConsultarLocalidades(cod_provincia)
            print("\n".join(ret))

        if '--bonificaciones_penalizaciones' in sys.argv:
            ret = wslum.ConsultarBonificacionesPenalizaciones()
            print("\n".join(ret))

        if '--otros_impuestos' in sys.argv:
            ret = wslum.ConsultarOtrosImpuestos()
            print("\n".join(ret))

        if '--puntosventa' in sys.argv:
            ret = wslum.ConsultarPuntosVentas()
            print("\n".join(ret))

        print("hecho.")

    except SoapFault as e:
        print("Falla SOAP:", e.faultcode, e.faultstring.encode("ascii", "ignore"), file=sys.stderr)
        sys.exit(3)
    except Exception as e:
        try:
            print(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[0], file=sys.stderr)
        except BaseException:
            print("Excepción no disponible:", type(e), file=sys.stderr)
        if DEBUG:
            raise
        sys.exit(5)
    finally:
        if XML:
            open("wslum_request.xml", "w").write(wslum.client.xml_request)
            open("wslum_response.xml", "w").write(wslum.client.xml_response)



#!/usr/bin/python
# -*- coding: utf8 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Herramienta para procesar y consultar el Padrón Unico de Contribuyentes AFIP"

# Documentación e información adicional:
#    http://www.sistemasagiles.com.ar/trac/wiki/PadronContribuyentesAFIP

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2014-2016 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.07e"


import csv
import json
import os
import shelve
import socket
import sqlite3
import urllib.request
import urllib.error
import urllib.parse
import zipfile
from email.utils import formatdate
import sys
import warnings
from .utils import leer, escribir, N, A, I, get_install_dir, safe_console, \
    inicializar_y_capturar_excepciones_simple, WebClient, norm, \
    exception_info


# formato y ubicación archivo completo de la condición tributaria según RG 1817

FORMATO = [
    ("nro_doc", 11, N, ""),
    ("denominacion", 30, A, ""),
    ("imp_ganancias", 2, A, "'NI', 'AC','EX', 'NC'"),
    ("imp_iva", 2, A, "'NI' , 'AC','EX','NA','XN','AN'"),
    ("monotributo", 2, A, "'NI', 'Codigo categoria tributaria'"),
    ("integrante_soc", 1, A, "'N' , 'S'"),
    ("empleador", 1, A, "'N', 'S'"),
    ("actividad_monotributo", 2, A, ""),
    ("tipo_doc", 2, N, "80: CUIT, 96: DNI, etc."),
    ("cat_iva", 2, N, "1: RI, 4: EX, 5: CF, 6: MT, etc"),
    ("email", 250, A, ""),
]

# Mapeos constantes:

PROVINCIAS = {0: 'CIUDAD AUTONOMA BUENOS AIRES', 1: 'BUENOS AIRES',
              2: 'CATAMARCA', 3: 'CORDOBA', 4: 'CORRIENTES', 5: 'ENTRE RIOS', 6: 'JUJUY',
              7: 'MENDOZA', 8: 'LA RIOJA', 9: 'SALTA', 10: 'SAN JUAN', 11: 'SAN LUIS',
              12: 'SANTA FE', 13: 'SANTIAGO DEL ESTERO', 14: 'TUCUMAN', 16: 'CHACO',
              17: 'CHUBUT', 18: 'FORMOSA', 19: 'MISIONES', 20: 'NEUQUEN', 21: 'LA PAMPA',
              22: 'RIO NEGRO', 23: 'SANTA CRUZ', 24: 'TIERRA DEL FUEGO'}

TIPO_CLAVE = {'CUIT': 80, 'CUIL': 86, 'CDI': 86, 'DNI': 96, 'Otro': 99}

DEBUG = True

URL = "http://www.afip.gob.ar/genericos/cInscripcion/archivos/apellidoNombreDenominacion.zip"
URL_API = "https://soa.afip.gob.ar/"


class PadronAFIP():
    "Interfaz para consultar situación tributaria (Constancia de Inscripcion)"

    _public_methods_ = ['Buscar', 'Descargar', 'Procesar', 'Guardar',
                        'ConsultarDomicilios', 'Consultar', 'Conectar',
                        'DescargarConstancia', 'MostrarPDF',
                        "ObtenerTablaParametros",
                        ]
    _public_attrs_ = ['InstallDir', 'Traceback', 'Excepcion', 'Version',
                      'cuit', 'dni', 'denominacion', 'imp_ganancias', 'imp_iva',
                      'monotributo', 'integrante_soc', 'empleador',
                      'actividad_monotributo', 'cat_iva', 'domicilios',
                      'tipo_doc', 'nro_doc', 'LanzarExcepciones',
                      'tipo_persona', 'estado', 'impuestos', 'actividades',
                      'direccion', 'localidad', 'provincia', 'cod_postal',
                      'data', 'response',
                      ]
    _readonly_attrs_ = _public_attrs_[3:-1]
    _reg_progid_ = "PadronAFIP"
    _reg_clsid_ = "{6206DF5E-3EEF-47E9-A532-CD81EBBAF3AA}"

    def __init__(self):
        self.db_path = os.path.join(self.InstallDir, "padron.db")
        self.Version = __version__
        # Abrir la base de datos
        self.db = sqlite3.connect(self.db_path)
        self.db.row_factory = sqlite3.Row
        self.cursor = self.db.cursor()
        self.LanzarExcepciones = False
        self.inicializar()
        self.client = None

    def inicializar(self):
        self.Excepcion = self.Traceback = ""
        self.cuit = self.dni = 0
        self.tipo_persona = ""                      # FISICA o JURIDICA
        self.tipo_doc = 0
        self.estado = ""                            # ACTIVO
        self.denominacion = ""
        self.direccion = self.localidad = self.provincia = self.cod_postal = ""
        self.domicilios = []
        self.impuestos = []
        self.actividades = []
        self.imp_iva = self.empleador = self.integrante_soc = self.cat_iva = ""
        self.monotributo = self.actividad_monotributo = ""
        self.data = {}
        self.response = ""

    @inicializar_y_capturar_excepciones_simple
    def Conectar(self, url=URL_API, proxy="", wrapper=None, cacert=None, trace=False):
        self.client = WebClient(location=url, trace=trace, cacert=cacert)
        self.client.method = "GET"       # metodo RESTful predeterminado
        self.client.enctype = None       # no enviar body
        return True

    @inicializar_y_capturar_excepciones_simple
    def Descargar(self, url=URL, filename="padron.txt", proxy=None):
        "Descarga el archivo de AFIP, devuelve 200 o 304 si no fue modificado"
        proxies = {}
        if proxy:
            proxies['http'] = proxy
            proxies['https'] = proxy
            proxy_handler = urllib.request.ProxyHandler(proxies)
        print("Abriendo URL %s ..." % url)
        req = urllib.request.Request(url)
        if os.path.exists(filename):
            http_date = formatdate(timeval=os.path.getmtime(filename),
                                   localtime=False, usegmt=True)
            req.add_header('If-Modified-Since', http_date)
        try:
            web = urllib.request.urlopen(req)
        except urllib.error.HTTPError as e:
            if e.code == 304:
                print("No modificado desde", http_date)
                return 304
            else:
                raise
        # leer info del request:
        meta = web.info()
        lenght = float(meta['Content-Length'])
        date = meta['Last-Modified']
        tmp = open(filename + ".zip", "wb")
        print("Guardando")
        size = 0
        p0 = None
        while True:
            p = int(size / lenght * 100)
            if p0 is None or p > p0:
                print("Leyendo ... %0d %%" % p)
                p0 = p
            data = web.read(1024 * 100)
            size = size + len(data)
            if not data:
                print("Descarga Terminada!")
                break
            tmp.write(data)
        print("Abriendo ZIP...")
        tmp.close()
        web.close()
        uf = open(filename + ".zip", "rb")
        zf = zipfile.ZipFile(uf)
        for fn in zf.namelist():
            print("descomprimiendo", fn)
            tf = open(filename, "wb")
            tf.write(zf.read(fn))
            tf.close()
        return 200

    @inicializar_y_capturar_excepciones_simple
    def Procesar(self, filename="padron.txt", borrar=False):
        "Analiza y crea la base de datos interna sqlite para consultas"
        f = open(filename, "r")
        keys = [k for k, l, t, d in FORMATO]
        # conversion a planilla csv (no usado)
        if False and not os.path.exists("padron.csv"):
            csvfile = open('padron.csv', 'wb')
            import csv
            wr = csv.writer(csvfile, delimiter=',',
                            quotechar='"', quoting=csv.QUOTE_MINIMAL)
            for i, l in enumerate(f):
                if i % 100000 == 0:
                    print("Progreso: %d registros" % i)
                r = leer(l, FORMATO)
                row = [r[k] for k in keys]
                wr.writerow(row)
            csvfile.close()
        f.seek(0)
        if os.path.exists(self.db_path) and borrar:
            os.remove(self.db_path)
        if True:
            db = db = sqlite3.connect(self.db_path)
            c = db.cursor()
            c.execute("CREATE TABLE padron ("
                      "nro_doc INTEGER, "
                      "denominacion VARCHAR(30), "
                      "imp_ganancias VARCHAR(2), "
                      "imp_iva VARCHAR(2), "
                      "monotributo VARCHAR(1), "
                      "integrante_soc VARCHAR(1), "
                      "empleador VARCHAR(1), "
                      "actividad_monotributo VARCHAR(2), "
                      "tipo_doc INTEGER, "
                      "cat_iva INTEGER DEFAULT NULL, "
                      "email VARCHAR(250), "
                      "PRIMARY KEY (tipo_doc, nro_doc)"
                      ");")
            c.execute("CREATE TABLE domicilio ("
                      "id INTEGER PRIMARY KEY AUTOINCREMENT, "
                      "tipo_doc INTEGER, "
                      "nro_doc INTEGER, "
                      "direccion TEXT, "
                      "FOREIGN KEY (tipo_doc, nro_doc) REFERENCES padron "
                      ");")
            # importar los datos a la base sqlite
            for i, l in enumerate(f):
                if i % 10000 == 0:
                    print(i)
                l = l.strip("\x00")
                r = leer(l, FORMATO)
                params = [r[k] for k in keys]
                params[8] = 80          # agrego tipo_doc = CUIT
                params[9] = None        # cat_iva no viene de AFIP
                placeholders = ", ".join(["?"] * len(params))
                c.execute("INSERT INTO padron VALUES (%s)" % placeholders,
                          params)
            db.commit()
            c.close()
            db.close()

    @inicializar_y_capturar_excepciones_simple
    def Buscar(self, nro_doc, tipo_doc=80):
        "Devuelve True si fue encontrado y establece atributos con datos"
        # cuit: codigo único de identificación tributaria del contribuyente
        #       (sin guiones)
        self.cursor.execute("SELECT * FROM padron WHERE "
                            " tipo_doc=? AND nro_doc=?", [tipo_doc, nro_doc])
        row = self.cursor.fetchone()
        for key in [k for k, l, t, d in FORMATO]:
            if row:
                val = row[key]
                if not isinstance(val, str):
                    val = str(row[key])
                setattr(self, key, val)
            else:
                setattr(self, key, '')
        if self.tipo_doc == 80:
            self.cuit = self.nro_doc
        elif self.tipo_doc == 96:
            self.dni = self.nro_doc
        # determinar categoría de IVA (tentativa)
        try:
            cat_iva = int(self.cat_iva)
        except ValueError:
            cat_iva = None
        if cat_iva:
            pass
        elif self.imp_iva in ('AC', 'S'):
            self.cat_iva = 1  # RI
        elif self.imp_iva == 'EX':
            self.cat_iva = 4  # EX
        elif self.monotributo:
            self.cat_iva = 6  # MT
        else:
            self.cat_iva = 5  # CF
        return True if row else False

    @inicializar_y_capturar_excepciones_simple
    def ConsultarDomicilios(self, nro_doc, tipo_doc=80, cat_iva=None):
        "Busca los domicilios, devuelve la cantidad y establece la lista"
        self.cursor.execute("SELECT direccion FROM domicilio WHERE "
                            " tipo_doc=? AND nro_doc=? ORDER BY id ",
                            [tipo_doc, nro_doc])
        filas = self.cursor.fetchall()
        self.domicilios = [fila['direccion'] for fila in filas]
        return len(filas)

    @inicializar_y_capturar_excepciones_simple
    def Guardar(self, tipo_doc, nro_doc, denominacion, cat_iva, direccion,
                email, imp_ganancias='NI', imp_iva='NI', monotributo='NI',
                integrante_soc='N', empleador='N'):
        "Agregar o actualizar los datos del cliente"
        if self.Buscar(nro_doc, tipo_doc):
            sql = ("UPDATE padron SET denominacion=?, cat_iva=?, email=?, "
                   "imp_ganancias=?, imp_iva=?, monotributo=?, "
                   "integrante_soc=?, empleador=? "
                   "WHERE tipo_doc=? AND nro_doc=?")
            params = [denominacion, cat_iva, email, imp_ganancias,
                      imp_iva, monotributo, integrante_soc, empleador,
                      tipo_doc, nro_doc]
        else:
            sql = ("INSERT INTO padron (tipo_doc, nro_doc, denominacion, "
                   "cat_iva, email, imp_ganancias, imp_iva, monotributo, "
                   "integrante_soc, empleador) "
                   "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)")
            params = [tipo_doc, nro_doc, denominacion, cat_iva, email,
                      imp_ganancias, imp_iva, monotributo,
                      integrante_soc, empleador]
        self.cursor.execute(sql, params)
        # agregar el domicilio solo si no existe:
        if direccion:
            self.cursor.execute("SELECT * FROM domicilio WHERE direccion=? "
                                "AND tipo_doc=? AND nro_doc=?",
                                [direccion, tipo_doc, nro_doc])
            if self.cursor.rowcount < 0:
                sql = ("INSERT INTO domicilio (nro_doc, tipo_doc, direccion)"
                       "VALUES (?, ?, ?)")
                self.cursor.execute(sql, [nro_doc, tipo_doc, direccion])
        self.db.commit()
        return True

    @inicializar_y_capturar_excepciones_simple
    def Consultar(self, nro_doc):
        "Llama a la API pública de AFIP para obtener los datos de una persona"
        n = 0
        while n <= 4:
            n += 1                          # reintentar 3 veces
            try:
                if not self.client:
                    if DEBUG:
                        warnings.warn("reconectando intento [%d]..." % n)
                    self.Conectar()
                self.response = self.client("sr-padron", "v2", "persona", str(nro_doc))
            except Exception as e:
                self.client = None
                ex = exception_info()
                self.Traceback = ex.get("tb", "")
                try:
                    self.Excepcion = norm(ex.get("msg", "").replace("\n", ""))
                except BaseException:
                    self.Excepcion = "<no disponible>"
                if DEBUG:
                    warnings.warn("Error %s [%d]" % (self.Excepcion, n))
            else:
                break
        else:
            return False
        result = json.loads(self.response)
        if result['success']:
            data = result['data']
            # extraigo datos generales del contribuyente:
            self.cuit = data["idPersona"]
            self.tipo_persona = data["tipoPersona"]
            self.tipo_doc = TIPO_CLAVE.get(data["tipoClave"])
            self.dni = data.get("numeroDocumento")
            self.estado = data.get("estadoClave")
            self.denominacion = data.get("nombre")
            # analizo el domicilio
            domicilio = data.get("domicilioFiscal")
            if domicilio:
                self.direccion = domicilio.get("direccion", "")
                self.localidad = domicilio.get("localidad", "")  # no usado en CABA
                self.provincia = PROVINCIAS.get(domicilio.get("idProvincia"), "")
                self.cod_postal = domicilio.get("codPostal")
            else:
                self.direccion = self.localidad = self.provincia = ""
                self.cod_postal = ""
            # retrocompatibilidad:
            self.domicilios = ["%s - %s (%s) - %s" % (
                self.direccion, self.localidad,
                self.cod_postal, self.provincia,)]
            # analizo impuestos:
            self.impuestos = data.get("impuestos", [])
            self.actividades = data.get("actividades", [])
            if 32 in self.impuestos:
                self.imp_iva = "EX"
            elif 33 in self.impuestos:
                self.imp_iva = "NI"
            elif 34 in self.impuestos:
                self.imp_iva = "NA"
            else:
                self.imp_iva = "S" if 30 in self.impuestos else "N"
            mt = data.get("categoriasMonotributo", {})
            self.monotributo = "S" if mt else "N"
            self.actividad_monotributo = ""  # TODO: mt[0].get("idCategoria")
            self.integrante_soc = ""
            self.empleador = "S" if 301 in self.impuestos else "N"
            self.cat_iva = ""
            self.data = data
        else:
            error = result['error']
            self.Excepcion = error['mensaje']
        return True

    @inicializar_y_capturar_excepciones_simple
    def DescargarConstancia(self, nro_doc, filename="constancia.pdf"):
        "Llama a la API para descargar una constancia de inscripcion (PDF)"
        if not self.client:
            self.Conectar()
        self.response = self.client("sr-padron", "v1", "constancia", str(nro_doc))
        if self.response.startswith("{"):
            result = json.loads(self.response)
            assert not result["success"]
            self.Excepcion = result['error']['mensaje']
            return False
        else:
            with open(filename, "wb") as f:
                f.write(self.response)
            return True

    @inicializar_y_capturar_excepciones_simple
    def MostrarPDF(self, archivo, imprimir=False):
        if sys.platform.startswith(("linux2", 'java')):
            os.system("evince ""%s""" % archivo)
        else:
            operation = imprimir and "print" or ""
            os.startfile(archivo, operation)
        return True

    @inicializar_y_capturar_excepciones_simple
    def ObtenerTablaParametros(self, tipo_recurso, sep="||"):
        "Devuelve un array de elementos que tienen id y descripción"
        if not self.client:
            self.Conectar()
        self.response = self.client("parametros", "v1", tipo_recurso)
        result = json.loads(self.response.decode('utf8'))
        ret = {}
        if result['success']:
            data = result['data']
            # armo un diccionario con los datos devueltos:
            key = [k for k in list(data[0].keys()) if k.startswith("id")][0]
            val = [k for k in list(data[0].keys()) if k.startswith("desc")][0]
            for it in data:
                ret[it[key]] = it[val]
            self.data = data
        else:
            error = result['error']
            self.Excepcion = error['mensaje']
        if sep:
            return ["%s%%s%s%%s%s" % (sep, sep, sep) % it for it in sorted(ret.items())]
        else:
            return ret


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = PadronAFIP.InstallDir = get_install_dir()

if __name__ == "__main__":

    safe_console()

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(PadronAFIP)
    else:
        padron = PadronAFIP()
        padron.LanzarExcepciones = True
        import time
        t0 = time.time()
        if "--descargar" in sys.argv:
            padron.Descargar()
        if "--procesar" in sys.argv:
            padron.Procesar(borrar='--borrar' in sys.argv)
        if "--parametros" in sys.argv:
            import codecs
            import locale
            import traceback
            if sys.stdout.encoding is None:
                sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout, "replace")
                sys.stderr = codecs.getwriter(locale.getpreferredencoding())(sys.stderr, "replace")
            print("=== Impuestos ===")
            print('\n'.join(padron.ObtenerTablaParametros("impuestos")))
            print("=== Conceptos ===")
            print('\n'.join(padron.ObtenerTablaParametros("conceptos")))
            print("=== Actividades ===")
            print('\n'.join(padron.ObtenerTablaParametros("actividades")))
            print("=== Caracterizaciones ===")
            print('\n'.join(padron.ObtenerTablaParametros("caracterizaciones")))
            print("=== Categorias Monotributo ===")
            print('\n'.join(padron.ObtenerTablaParametros("categoriasMonotributo")))
            print("=== Categorias Autonomos ===")
            print('\n'.join(padron.ObtenerTablaParametros("categoriasAutonomo")))

        if '--csv' in sys.argv:
            csv_reader = csv.reader(open("entrada.csv", "rU"),
                                    dialect='excel', delimiter=",")
            csv_writer = csv.writer(open("salida.csv", "w"),
                                    dialect='excel', delimiter=",")
            encabezado = next(csv_reader)
            columnas = ["cuit", "denominacion", "estado", "direccion",
                        "localidad", "provincia", "cod_postal",
                        "impuestos", "actividades", "imp_iva",
                        "monotributo", "actividad_monotributo",
                        "empleador", "imp_ganancias", "integrante_soc"]
            csv_writer.writerow(columnas)

            for fila in csv_reader:
                cuit = (fila[0] if fila else "").replace("-", "")
                if cuit.isdigit():
                    if '--online' in sys.argv:
                        padron.Conectar(trace="--trace" in sys.argv)
                        print("Consultando AFIP online...", cuit, end=' ')
                        ok = padron.Consultar(cuit)
                    else:
                        print("Consultando AFIP local...", cuit, end=' ')
                        ok = padron.Buscar(cuit)
                    print('ok' if ok else "error", padron.Excepcion)
                    # domicilio posiblemente esté en Latin1, normalizar
                    csv_writer.writerow([norm(getattr(padron, campo, ""))
                                         for campo in columnas])
        elif "--reconex" in sys.argv:
            padron.Conectar(trace="--trace" in sys.argv)
            cuit = 20267565393
            for i in range(10000):
                t0 = time.time()
                ok = padron.Consultar(cuit)
                t1 = time.time()
                print(("%2.4f" % (t1 - t0)))
        else:
            cuit = len(sys.argv) > 1 and sys.argv[1] or "20267565393"
            # consultar un cuit:
            if '--online' in sys.argv:
                padron.Conectar(trace="--trace" in sys.argv)
                print("Consultando AFIP online...", end=' ')
                ok = padron.Consultar(cuit)
                print('ok' if ok else "error", padron.Excepcion)
                print("Denominacion:", padron.denominacion)
                print("CUIT:", padron.cuit)
                print("Tipo:", padron.tipo_persona, padron.tipo_doc, padron.dni)
                print("Estado:", padron.estado)
                print("Direccion:", padron.direccion)
                print("Localidad:", padron.localidad)
                print("Provincia:", padron.provincia)
                print("Codigo Postal:", padron.cod_postal)
                print("Impuestos:", padron.impuestos)
                print("Actividades:", padron.actividades)
                print("IVA", padron.imp_iva)
                print("MT", padron.monotributo, padron.actividad_monotributo)
                print("Empleador", padron.empleador)
            elif '--constancia' in sys.argv:
                filename = sys.argv[2]
                print("Descargando constancia AFIP online...", cuit, filename)
                ok = padron.DescargarConstancia(cuit, filename)
                print('ok' if ok else "error", padron.Excepcion)
                if '--mostrar' in sys.argv:
                    padron.MostrarPDF(archivo=filename,
                                      imprimir='--imprimir' in sys.argv)
            else:
                ok = padron.Buscar(cuit)
                if ok:
                    print("Denominacion:", padron.denominacion)
                    print("IVA:", padron.imp_iva)
                    print("Ganancias:", padron.imp_ganancias)
                    print("Monotributo:", padron.monotributo)
                    print("Integrante Soc.:", padron.integrante_soc)
                    print("Empleador", padron.empleador)
                    print("Actividad Monotributo:", padron.actividad_monotributo)
                    print("Categoria IVA:", padron.cat_iva)
                    padron.ConsultarDomicilios(cuit)
                    for dom in padron.domicilios:
                        print(dom)
                else:
                    print(padron.Excepcion)
                    print(padron.Traceback)
        t1 = time.time()
        if '--trace' in sys.argv:
            print("tiempo", t1 - t0)
